from collections import namedtuple

import h5py

from ...transforms import StaticTransform
from ...utils import recursively_load_dict_contents_from_group
from ..base import TransformDataset, default_image_datapoint


class StaticSet(TransformDataset):
    def __init__(self, *data_keys, transforms=None):
        """
        Abstract class for static datasets. Defines data_keys and a corresponding datapoint.
        """
        super().__init__(transforms=transforms)

        self.data_keys = data_keys
        if set(data_keys) == {"images", "responses"}:
            # this version IS serializable in pickle
            self.data_point = default_image_datapoint
        else:
            # this version is NOT - you cannot use this with a dataloader with num_workers > 1
            self.data_point = namedtuple("DataPoint", data_keys)


class H5ArraySet(StaticSet):
    def __init__(self, filename, *data_keys, transforms=None):
        """
        Dataset for static data stored in hdf5 files.
        Args:
            filename:      filename of the hdf5 file
            *data_keys:    data keys to be read from the file
            transforms:    list of transforms applied to each datapoint
        """
        super().__init__(*data_keys, transforms=transforms)

        self._fid = h5py.File(filename, "r")
        self.data = self._fid
        self.data_loaded = False
        m = None
        for key in data_keys:
            assert key in self.data, "Could not find {} in file".format(key)
            if m is None:
                m = len(self.data[key])
            else:
                assert m == len(self.data[key]), "Length of datasets do not match"
        self._len = m

    def load_content(self):
        self.data = recursively_load_dict_contents_from_group(self._fid)
        self.data_loaded = True

    def unload_content(self):
        self.data = self._fid
        self.data_loaded = False

    def __getitem__(self, item):
        x = self.data_point(*(self.data[g][item] for g in self.data_keys))
        for tr in self.transforms:
            assert isinstance(tr, StaticTransform)
            x = tr(x)
        return x

    def __iter__(self):
        yield from map(self.__getitem__, range(len(self)))

    def __len__(self):
        return self._len

    def __repr__(self):
        return "\n".join(
            ["Tensor {}: {} ".format(key, self.data[key].shape) for key in self.data_keys]
            + ["Transforms: " + repr(self.transforms)]
        )

    def __getattr__(self, item):
        if item in self.data:
            item = self.data[item]
            if isinstance(item, h5py.Dataset):
                dtype = item.dtype
                item = item[()]
                if dtype.char == "S":  # convert bytes to univcode
                    item = item.astype(str)
                return item
            return item
        else:
            raise AttributeError("Item {} not found in {}".format(item, self.__class__.__name__))


class StaticImageSet(H5ArraySet):
    def __init__(self, filename, *data_keys, transforms=None, cache_raw=False, stats_source=None):
        """
        Dataset for h5 files.
        Args:
            filename:       filename of the hdf5 file
            *data_keys:     datasets to be extracted
            transforms:     transforms applied to each data point
            cache_raw:      whether to cache the raw (untransformed) datapoints
            stats_source:   statistic source to be used.
        """
        super().__init__(filename, *data_keys, transforms=transforms)
        self.cache_raw = cache_raw
        self.last_raw = None
        self.stats_source = stats_source if stats_source is not None else "all"

    @property
    def n_neurons(self):
        return len(self[0].responses)

    @property
    def neurons(self):
        return AttributeTransformer("neurons", self.data, self.transforms, data_group="responses")

    @property
    def info(self):
        return AttributeHandler("item_info", self.data)

    @property
    def img_shape(self):
        return (1,) + self[0].images.shape

    def transformed_mean(self, stats_source=None):
        if stats_source is None:
            stats_source = self.stats_source

        tmp = [np.atleast_1d(self.statistics[dk][stats_source]["mean"][()]) for dk in self.data_keys]
        return self.transform(self.data_point(*tmp))

    def __repr__(self):
        return super().__repr__() + (
            "\n\t[Stats source: {}]".format(self.stats_source) if self.stats_source is not None else ""
        )

    def __dir__(self):
        attrs = set(self.__dict__).union(set(dir(type(self))))
        return attrs.union(set(self.data.keys()))



"""
This package contains all tools that are useful for training system identification models
as well as things that can be applied to neural network training in general.
This includes:
  - context_managers: managing device state and eval state
  - cyclers: objects for cycling through the data
  - early_stopping: controlling the training loop
  - tracking: objects that can be used for training performance and progress during training
"""

from .context_managers import device_state, eval_state
from .cyclers import Exhauster, LongCycler, ShortCycler
from .early_stopping import early_stopping
from .tracking import MultipleObjectiveTracker, TimeObjectiveTracker



"""
This package contains shifter-networks. These are used to modify the neurons' readout position
based on the eye position.

Usually, the input to the shifters are the pupil centers of both eyes. Given that input,
a horizontal and vertical shift of the readout position is returned by the shifter.
"""
from .base import Shifter
from .mlp import MLP, MLPShifter
from .static_affine import StaticAffine2d, StaticAffine2dShifter



import torch
import featurevis
import featurevis.ops as ops
from featurevis.utils import Compose
from .mask import create_mask
from .general import SingleNeuronModel, infer_device


def generate_mei(
    model,
    input_shape,
    pixel_min,
    pixel_max,
    std=None,
    norm=None,
    gb=None,
    zscore=0.3,
    step_size=10,
    num_iterations=1000,
    print_iters=1001,
):
    model_items = list(model.parameters()) + list(model.buffers())
    device = model_items[0].device

    if std is None and norm is None:
        raise ValueError("Either std or norm should be provided.")
    elif std is not None and norm is not None:
        raise ValueError("Only one of std or norm should be provided.")

    mean = (pixel_min + pixel_max) / 2

    # get image to optimize
    initial_image = torch.randn(1, *input_shape, dtype=torch.float32).to(device) + mean

    if std is not None:
        change_stats_transform = ops.ChangeStats(std=std, mean=mean)
    elif norm is not None:
        change_stats_transform = ops.ChangeNorm(norm=norm, mean=mean)

    post_update = Compose([change_stats_transform, ops.ClipRange(pixel_min, pixel_max)])

    initial_image = post_update(initial_image)

    if gb is not None:
        gb = ops.GaussianBlur(gb)

    # optimization of the image
    mei, act, reg_values = featurevis.gradient_ascent(
        model,
        initial_image,
        step_size=step_size,
        num_iterations=num_iterations,
        post_update=post_update,
        gradient_f=gb,
        print_iters=print_iters,
    )

    # Grab the activation of the last optimization step
    mei_act = act[-1]

    # remove batch dimension
    mei = mei[0]
    channels, height, width = mei.shape

    # get mask and it's center from mei (change zscore_thresh to change mask size)
    # mask, mask_center = create_mask(mei, return_mask_centroid=True, zscore_thresh=.5)
    mask, mask_center = create_mask(
        mei.mean(dim=0), return_mask_centroid=True, zscore_thresh=zscore
    )

    return mei.cpu().data.numpy(), mei_act, mask.cpu().data.numpy(), mask_center


def get_mei_dict(
    response_predicting_model,
    input_shape,
    pixel_value_lower_bound,
    pixel_value_upper_bound,
    neuron_idxs=None,
    required_img_std=None,
    required_img_norm=None,
    gb=None,
    zscore=0.3,  # TODO: think more about this (either user needs to specify this or we need to find a way to calculate this based on other params)
    step_size=10,
    num_iterations=1000,
    print_iters=1001,
    device=None,
):
    if device is None:
        device = infer_device(response_predicting_model)

    if neuron_idxs is None:
        initial_image = torch.randn(1, *input_shape, dtype=torch.float32).to(device)
        temp_out = response_predicting_model(initial_image)
        neuron_idxs = list(range(temp_out.shape[1]))

    meis_dict = {}
    for idx, neuron_idx in enumerate(neuron_idxs):
        print(f"neuron_idx = {neuron_idx}: neuron number {idx}/{len(neuron_idxs)}")
        model = SingleNeuronModel(response_predicting_model, neuron_idx).to(device)
        mei, mei_act, mask, mask_center = generate_mei(
            model,
            input_shape,
            std=required_img_std,
            norm=required_img_norm,
            pixel_min=pixel_value_lower_bound,
            pixel_max=pixel_value_upper_bound,
            gb=gb,
            zscore=zscore,
            step_size=step_size,
            num_iterations=num_iterations,
            print_iters=print_iters,
        )
        meis_dict[idx] = {
            "mei": mei,
            "activation": mei_act,
            "mask": mask,
            "center_pos": mask_center,
        }
    return meis_dict


def check_meis(
    meis_dict,
    required_pixel_min,
    required_pixel_max,
    required_std=None,
    required_norm=None,
):
    return True



from abc import ABC, abstractmethod

from torch import nn


class Core(ABC):
    """
    Base class for the core models, taking 2d inputs and computing nonlinear features.
    """

    def initialize(self):
        """
        Initialization applied on the core.
        """
        self.apply(self.init_conv)

    @staticmethod
    def init_conv(m):
        """
        Initialize convolution layers with:
            - weights: xavier_normal
            - biases: 0

        Args:
            m (nn.Module): a pytorch nn module.
        """
        if isinstance(m, nn.Conv2d):
            nn.init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0)

    @abstractmethod
    def regularizer(self):
        """
        Regularization applied on the core. Returns a scalar value.
        """

    @abstractmethod
    def forward(self, x):
        """
        Forward function for pytorch nn module.

        Args:
            x (torch.tensor): input of shape (batch, channels, height, width)
        """

    def __repr__(self):
        s = super().__repr__()
        s += f" [{self.__class__.__name__} regularizers: "
        ret = []
        for attr in filter(lambda x: "gamma" in x or "skip" in x, dir(self)):
            ret.append(f"{attr} = {getattr(self, attr)}")
        return s + "|".join(ret) + "]\n"



from torch import nn


class Shifter(nn.Module):
    """
    Abstract base class for a shifter. It's strongly adviced that the regularizer and initialize methods are implemented by the inheriting class.
    """

    def __repr__(self):
        s = super().__repr__()
        s += " [{} regularizers: ".format(self.__class__.__name__)
        ret = []
        for attr in filter(lambda x: "gamma" in x, dir(self)):
            ret.append("{} = {}".format(attr, getattr(self, attr)))
        return s + "|".join(ret) + "]\n"

    def regularizer(self):
        """
        Regularizer method to be used during training.
        """
        raise NotImplementedError("regularizer method must be implemented by the inheriting class")

    def initialize(self):
        """
        weight initialization of the torch.parameters
        """
        raise NotImplementedError("initialize method must be implemented by the inheriting class")



import torch


def varargin(f):
    """ Decorator to make a function able to ignore named parameters not declared in its
     definition.

    Arguments:
        f (function): Original function.

    Usage:
            @varargin
            def my_f(x):
                # ...
        is equivalent to
            def my_f(x, **kwargs):
                #...
        Using the decorator is recommended because it makes it explicit that my_f won't
        use arguments received in the kwargs dictionary.
    """
    import inspect
    import functools

    # Find the name of parameters expected by f
    f_params = inspect.signature(f).parameters.values()
    param_names = [p.name for p in f_params]  # name of parameters expected by f
    receives_kwargs = any([p.kind == inspect.Parameter.VAR_KEYWORD for p in
                           f_params])  # f receives a dictionary of **kwargs

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if not receives_kwargs:
            # Ignore named parameters not expected by f
            kwargs = {k: kwargs[k] for k in kwargs.keys() if k in param_names}
        return f(*args, **kwargs)

    return wrapper


class Compose():
    """ Chain a set of operations into a single function.

    Each function must receive one positional argument and any number of keyword
    arguments. Each function is called with the output of the previous one (as its
    positional argument) and all keyword arguments of the original call.

    Arguments:
        operations (list): List of functions.
    """

    def __init__(self, operations):
        self.operations = operations

    def __call__(self, x, **kwargs):
        if len(self.operations) == 0:
            out = None
        else:
            out = self.operations[0](x, **kwargs)
            for op in self.operations[1:]:
                out = op(out, **kwargs)

        return out

    def __getitem__(self, item):
        return self.operations[item]


class Combine():
    """ Applies different operations to an input and combines its output.

    Arguments:
        operations (list): List of operations
        combine_op (function): Function used to combine the results of all the operations.
    """

    def __init__(self, operations, combine_op=torch.sum):
        self.operations = operations
        self.combine_op = combine_op

    def __call__(self, *args, **kwargs):
        if len(self.operations) == 0:
            return
        else:
            results = [op(*args, **kwargs) for op in self.operations]
            return self.combine_op(torch.stack(results, dim=0))

    def __getitem__(self, item):
        return self.operations[item]


import logging

logger = logging.getLogger(__name__)


def corr(y1, y2, dim=-1, eps=1e-8, **kwargs):
    """
    Compute the correlation between two PyTorch tensors along the specified dimension(s).

    Args:
        y1:      first PyTorch tensor
        y2:      second PyTorch tensor
        dim:     dimension(s) along which the correlation is computed. Any valid PyTorch dim spec works here
        eps:     offset to the standard deviation to avoid exploding the correlation due to small division (default 1e-8)
        **kwargs: passed to final numpy.mean operatoin over standardized y1 * y2

    Returns: correlation tensor
    """
    y1 = (y1 - y1.mean(dim=dim, keepdim=True)) / (y1.std(dim=dim, keepdim=True) + eps)
    y2 = (y2 - y2.mean(dim=dim, keepdim=True)) / (y2.std(dim=dim, keepdim=True) + eps)
    return (y1 * y2).mean(dim=dim, **kwargs)



import warnings

import torch
import torch.nn.functional as F
from scipy import signal

from featurevis.utils import varargin


################################## REGULARIZERS ##########################################
class TotalVariation():
    """ Total variation regularization.

    Arguments:
        weight (float): Weight of the regularization.
        isotropic (bool): Whether to use the isotropic or anisotropic definition of Total
            Variation. Default is anisotropic (l1-norm of the gradient).
    """
    def __init__(self, weight=1, isotropic=False):
        self.weight = weight
        self.isotropic = isotropic

    @varargin
    def __call__(self, x):
        # Using the definitions from Wikipedia.
        diffs_y = torch.abs(x[:, :, 1:] - x[:, :, -1:])
        diffs_x = torch.abs(x[:, :, :, 1:] - x[:, :, :, :-1])
        if self.isotropic:
            tv = torch.sqrt(diffs_y[:, :, :, :-1] ** 2 +
                            diffs_x[:, :, :-1, :] ** 2).reshape(len(x), -1).sum(-1)  # per image
        else:
            tv = diffs_y.reshape(len(x), -1).sum(-1) + diffs_x.reshape(len(x), -1).sum(-1)  # per image
        loss = self.weight * torch.mean(tv)

        return loss


class LpNorm():
    """Computes the lp-norm of an input.

    Arguments:
        weight (float): Weight of the regularization
        p (int): Degree for the l-p norm.
    """
    def __init__(self, weight=1, p=6):
        self.weight = weight
        self.p = p

    @varargin
    def __call__(self, x):
        lpnorm = (torch.abs(x) ** self.p).reshape(len(x), -1).sum(-1) ** (1 / self.p)
        loss = self.weight * torch.mean(lpnorm)
        return loss


class Similarity():
    """ Compute similarity metrics across all examples in one batch.

    Arguments:
        weight (float): Weight of the regularization.
        metric (str): What metric to use when computing pairwise similarities. One of:
            correlation: Masked correlation.
            cosine: Cosine similarity of the masked input.
            neg_euclidean: Negative of euclidean distance between the masked input.
        combine_op (function): Function used to agglomerate pairwise similarities.
        mask (torch.tensor or None): Mask to use when calculating similarities. Expected
            to be in [0, 1] range and be broadcastable with input.
    """
    def __init__(self, weight=1, metric='correlation', combine_op=torch.max, mask=None):
        self.weight = weight
        self.metric = metric
        self.combine_op = combine_op
        self.mask = mask

    @varargin
    def __call__(self, x):
        if len(x) < 2:
            warnings.warn('Only one image in the batch. Similarity regularization will'
                          'return 0')
            return 0

        # Mask x
        masked_x = x if self.mask is None else x * self.mask
        flat_x = masked_x.view(len(x), -1)

        # Compute similarity matrix
        if self.metric == 'correlation':
            if self.mask is None:
                residuals = flat_x - flat_x.mean(-1, keepdim=True)
                numer = torch.mm(residuals, residuals.t())
                ssr = (residuals ** 2).sum(-1)
            else:
                mask_sum = self.mask.sum() * (flat_x.shape[-1] / len(self.mask.view(-1)))
                mean = flat_x.sum(-1) / mask_sum
                residuals = x - mean.view(len(x), *[1, ] * (x.dim() - 1))  # N x 1 x 1 x 1
                numer = (residuals[None, :] * residuals[:, None] * self.mask).view(
                    len(x), len(x), -1).sum(-1)
                ssr = ((residuals ** 2) * self.mask).view(len(x), -1).sum(-1)
            sim_matrix = numer / (torch.sqrt(torch.ger(ssr, ssr)) + 1e-9)
        elif self.metric == 'cosine':
            norms = torch.norm(flat_x, dim=-1)
            sim_matrix = torch.mm(flat_x, flat_x.t()) / (torch.ger(norms, norms) + 1e-9)
        elif self.metric == 'neg_euclidean':
            sim_matrix = -torch.norm(flat_x[None, :, :] - flat_x[:, None, :], dim=-1)
        else:
            raise ValueError('Invalid metric name:{}'.format(self.metric))

        # Compute overall similarity
        triu_idx = torch.triu(torch.ones(len(x), len(x)), diagonal=1) == 1
        similarity = self.combine_op(sim_matrix[triu_idx])

        loss = self.weight * similarity

        return loss


# class PixelCNN():
#     def __init__(self, weight=1):
#         self.weight = weight
#
#         self.pixel_cnn = ... # load the model
#
#     @varargin
#     def __call__(self, x):
#         # Modify x to make it a valid input to pixel cnn (add channels)
#         prior = self.pixel_cnn(x)
#         loss = self.weight * prior


################################ TRANSFORMS ##############################################
class Jitter():
    """ Jitter the image at random by some certain amount.

    Arguments:
        max_jitter(tuple of ints): Maximum amount of jitter in y, x.
    """
    def __init__(self, max_jitter):
        self.max_jitter = max_jitter if isinstance(max_jitter, tuple) else (max_jitter,
                                                                            max_jitter)

    @varargin
    def __call__(self, x):
        # Sample how much to jitter
        jitter_y = torch.randint(-self.max_jitter[0], self.max_jitter[0] + 1, (1,),
                                 dtype=torch.int32).item()
        jitter_x = torch.randint(-self.max_jitter[1], self.max_jitter[1] + 1, (1,),
                                 dtype=torch.int32).item()

        # Pad and crop the rest
        pad_y = (jitter_y, 0) if jitter_y >= 0 else (0, -jitter_y)
        pad_x = (jitter_x, 0) if jitter_x >= 0 else (0, -jitter_x)
        padded_x = F.pad(x, pad=(*pad_x, *pad_y), mode='reflect')

        # Crop
        h, w = x.shape[-2:]
        jittered_x = padded_x[..., slice(0, h) if jitter_y > 0 else slice(-jitter_y, None),
                              slice(0,w) if jitter_x > 0 else slice(-jitter_x, None)]

        return jittered_x


class RandomCrop():
    """ Take a random crop of the input image.

    Arguments:
        height (int): Height of the crop.
        width (int): Width of the crop
    """
    def __init__(self, height, width):
        self.height = height
        self.width = width

    @varargin
    def __call__(self, x):
        crop_y = torch.randint(0, max(0, x.shape[-2] - self.height) + 1, (1,),
                               dtype=torch.int32).item()
        crop_x = torch.randint(0, max(0, x.shape[-1] - self.width) + 1, (1,),
                               dtype=torch.int32).item()
        cropped_x = x[..., crop_y: crop_y + self.height, crop_x: crop_x + self.width]

        return cropped_x


class BatchedCrops():
    """ Create a batch of crops of the original image.

    Arguments:
        height (int): Height of the crop
        width (int): Width of the crop
        step_size (int or tuple): Number of pixels in y, x to step for each crop.
        sigma (float or tuple): Sigma in y, x for the gaussian mask applied to each batch.
            None to avoid masking

    Note:
        Increasing the stride of every convolution to stride * step_size produces the same
        effect in a much more memory efficient way but it will be architecture dependent
        and may not play nice with the rest of transforms.
    """
    def __init__(self, height, width, step_size, sigma=None):
        self.height = height
        self.width = width
        self.step_size = step_size if isinstance(step_size, tuple) else (step_size,) * 2
        self.sigma = sigma if sigma is None or isinstance(sigma, tuple) else (sigma,) * 2

        # If needed, create gaussian mask
        if sigma is not None:
            y_gaussian = signal.gaussian(height, std=self.sigma[0])
            x_gaussian = signal.gaussian(width, std=self.sigma[1])
            self.mask = y_gaussian[:, None] * x_gaussian

    @varargin
    def __call__(self, x):
        if len(x) > 1:
            raise ValueError('x can only have one example.')
        if x.shape[-2] < self.height or x.shape[-1] < self.width:
            raise ValueError('x should be larger than the expected crop')

        # Take crops
        crops = []
        for i in range(0, x.shape[-2] - self.height + 1, self.step_size[0]):
            for j in range(0, x.shape[-1] - self.width + 1, self.step_size[1]):
                crops.append(x[..., i: i + self.height, j: j + self.width])
        crops = torch.cat(crops, dim=0)

        # Multiply by a gaussian mask if needed
        if self.sigma is not None:
            mask = torch.as_tensor(self.mask, device=crops.device, dtype=crops.dtype)
            crops = crops * mask

        return crops


class ChangeRange():
    """ This changes the range of x as follows:
        new_x = sigmoid(x) * (desired_max - desired_min) + desired_min

    Arguments:
        x_min (float or tensor): Minimum desired value for the output. If a tensor it
            needs to be broadcastable with x.
        x_max (float or tensor): Minimum desired value for the output. If a tensor it
            needs to be broadcastable with x.
    """
    def __init__(self, x_min, x_max):
        self.x_min = x_min
        self.x_max = x_max

    @varargin
    def __call__(self, x):
        new_x = torch.sigmoid(x) * (self.x_max - self.x_min) + self.x_min
        return new_x


class Resize():
    """ Resize images.

    Arguments:
        scale_factor (float): Factor to rescale the images:
            new_h, new_w = round(scale_factor * (old_h, old_w)).
        resize_method (str): 'nearest' or 'bilinear' interpolation.

    Note:
        This changes the dimensions of the image.
    """
    def __init__(self, scale_factor, resize_method='bilinear'):
        self.scale_factor = scale_factor
        self.resample_method = resize_method

    @varargin
    def __call__(self, x):
        new_height = int(round(x.shape[-2] * self.scale_factor))
        new_width = int(round(x.shape[-1] * self.scale_factor))
        return F.upsample(x, (new_height, new_width), mode=self.resize_method)


class GrayscaleToRGB():
    """ Transforms a single channel image into three channels (by copying the channel)."""
    @varargin
    def __call__(self, x):
        if x.dim() != 4 or x.shape[1] != 1:
            raise ValueError('Image is not grayscale!')

        return x.expand(-1, 3, -1, -1)


class Identity():
    """ Transform that returns the input as is."""
    @varargin
    def __call__(self, x):
        return x


############################## GRADIENT OPERATIONS #######################################

class ClipRange():
    """Clip the value of x to some specified range.

    Arguments:
        x_min (float): Lower valid value.
        x_max (float): Higher valid value.
    """
    def __init__(self, x_min, x_max):
        self.x_min = x_min
        self.x_max = x_max

    @varargin
    def __call__(self, x):
        return torch.clamp(x, self.x_min, self.x_max)


class FourierSmoothing():
    """ Smooth the input in the frequency domain.

    Image is transformed to fourier domain, power densities at i, j are multiplied by
    (1 - ||f||)**freq_exp where ||f|| = sqrt(f_i**2 + f_j**2) and the image is brought
    back to the spatial domain:
        new_x = ifft((1 - freqs) ** freq_exp * fft(x))

    Arguments:
        freq_exp (float): Exponent for the frequency mask. Higher numbers produce more
            smoothing.

    Note:
        Consider just using Gaussian blurring. Faster and easier to explain.
    """
    def __init__(self, freq_exp):
        self.freq_exp = freq_exp

    @varargin
    def __call__(self, x):
        # Create mask of frequencies (following np.fft.rfftfreq and np.fft.fftfreq docs)
        h, w = x.shape[-2:]
        freq_y = torch.cat([torch.arange((h - 1) // 2 + 1, dtype=torch.float32),
                            -torch.arange(h // 2, 0, -1, dtype=torch.float32)]) / h  # fftfreq
        freq_x = torch.arange(w // 2 + 1, dtype=torch.float32) / w  # rfftfreq
        yx_freq = torch.sqrt(freq_y[:, None] ** 2 + freq_x ** 2)

        # Create smoothing mask
        norm_freq = yx_freq * torch.sqrt(torch.tensor(2.0))  # 0-1
        mask = (1 - norm_freq) ** self.freq_exp

        # Smooth
        freq = torch.rfft(x, signal_ndim=2)  # same output as np.fft.rfft2
        mask = torch.as_tensor(mask, device=freq.device, dtype=freq.dtype).unsqueeze(-1)
        smooth = torch.irfft(freq * mask, signal_ndim=2, signal_sizes=x.shape[-2:])
        return smooth


class DivideByMeanOfAbsolute():
    """ Divides x by the mean of absolute x. """
    @varargin
    def __call__(self, x):
        return x / torch.abs(x).view(len(x), -1).mean(-1)


class MultiplyBy():
    """Multiply x by some constant.

    Arguments:
        const: Number x will be multiplied by
        decay_factor: Compute const every iteration as `const + decay_factor * (iteration
            - 1)`. Ignored if None.
    """
    def __init__(self, const, decay_factor=None):
        self.const = const
        self.decay_factor = decay_factor

    @varargin
    def __call__(self, x, iteration=None):
        if self.decay_factor is None:
            const = self.const
        else:
            const = self.const + self.decay_factor * (iteration - 1)

        return const * x


########################### POST UPDATE OPERATIONS #######################################
class GaussianBlur():
    """ Blur an image with a Gaussian window.

    Arguments:
        sigma (float or tuple): Standard deviation in y, x used for the gaussian blurring.
        decay_factor (float): Compute sigma every iteration as `sigma + decay_factor *
            (iteration - 1)`. Ignored if None.
        truncate (float): Gaussian window is truncated after this number of standard
            deviations to each side. Size of kernel = 8 * sigma + 1
        pad_mode (string): Mode for the padding used for the blurring. Valid values are:
            'constant', 'reflect' and 'replicate'
    """
    def __init__(self, sigma, decay_factor=None, truncate=4, pad_mode='reflect'):
        self.sigma = sigma if isinstance(sigma, tuple) else (sigma,) * 2
        self.decay_factor = decay_factor
        self.truncate = truncate
        self.pad_mode = pad_mode

    @varargin
    def __call__(self, x, iteration=None):
        num_channels = x.shape[1]

        # Update sigma if needed
        if self.decay_factor is None:
            sigma = self.sigma
        else:
            sigma = tuple(s + self.decay_factor * (iteration - 1) for s in self.sigma)

        # Define 1-d kernels to use for blurring
        y_halfsize = max(int(round(sigma[0] * self.truncate)), 1)
        y_gaussian = signal.gaussian(2 * y_halfsize + 1, std=sigma[0])
        x_halfsize = max(int(round(sigma[1] * self.truncate)), 1)
        x_gaussian = signal.gaussian(2 * x_halfsize + 1, std=sigma[1])
        y_gaussian = torch.as_tensor(y_gaussian, device=x.device, dtype=x.dtype)
        x_gaussian = torch.as_tensor(x_gaussian, device=x.device, dtype=x.dtype)

        # Blur
        padded_x = F.pad(x, pad=(x_halfsize, x_halfsize, y_halfsize, y_halfsize),
                         mode=self.pad_mode)
        blurred_x = F.conv2d(padded_x, y_gaussian.repeat(num_channels, 1, 1)[..., None],
                             groups=num_channels)
        blurred_x = F.conv2d(blurred_x, x_gaussian.repeat(num_channels, 1, 1, 1),
                             groups=num_channels)
        final_x = blurred_x / (y_gaussian.sum() * x_gaussian.sum())  # normalize

        return final_x


class ChangeStd():
    """ Change the standard deviation of input.

        Arguments:
        std (float or tensor): Desired std. If tensor, it should be the same length as x.
    """
    def __init__(self, std):
        self.std = std

    @varargin
    def __call__(self, x):
        x_std = torch.std(x.view(len(x), -1), dim=-1)
        fixed_std = x * (self.std / (x_std + 1e-9)).view(len(x), *[1, ] * (x.dim() - 1))
        return fixed_std

class ChangeStats():
    """ Change the mean and the standard deviation of input.

        Arguments:
        std (float or tensor): Desired std. If tensor, it should be the same length as x.
        std (float or tensor): Desired mean. If tensor, it should be the same length as x.
    """
    def __init__(self, std, mean):
        self.std = std
        self.mean = mean 

    @varargin
    def __call__(self, x):
        x_std = torch.std(x.view(len(x), -1), dim=-1)
        x_mean = torch.mean(x.view(len(x), -1), dim=-1)
        x_centered = x - x_mean
        fixed_std = x_centered * (self.std / (x_std + 1e-9)).view(len(x_centered), *[1, ] * (x_centered.dim() - 1))
        fixed_x = fixed_std + self.mean
        return fixed_x


class ChangeNorm():
    """ Change the norm of the input. If mean is specified norm is be computed
     as euclidean distance of input image from image with pixels set to that value

        Arguments:
        norm (float or tensor): Desired norm.
        mean (float or tensor): reference pixel values from which to compute distance.
    """
    def __init__(self, norm, mean=0):
        self.norm = norm
        self.mean = mean
        
    @varargin
    def __call__(self, x):
        x_centered = (x - self.mean).view(len(x), -1)
        x_norm2 = torch.sum(x_centered**2, dim=-1, keepdim=True)
        x_norm = torch.sqrt(x_norm2)
        x_renorm = x_centered/(x_norm + 1e-9) * self.norm + self.mean
        x_renorm = x_renorm.reshape(x.shape)
        return x_renorm

####################################### LOSS #############################################
class MSE():
    """ Compute MSE loss between x and a target specified at construction.
    
    Arguments:
        target (torch.tensor): Tensor to match. Broadcastable with expected input.
    """
    def __init__(self, target):
        self.target = target

    def __call__(self, x):
        return F.mse_loss(x, self.target)

class CosineSimilarity():
    """ Computes cosine similarity between x and a target specified at construction.
    
    Similarity is computed over the last axes and averaged across all axes (if any).
    
    Arguments:
        target (torch.Tensor): Tensor to match. Same shape as expected input.
    """
    def __init__(self, target):
        self.target = target

    @varargin
    def __call__(self, x):
        return F.cosine_similarity(x, self.target, dim=-1).mean()


class PoissonLogLikelihood():
    """ Computes (average) poisson log likelihood between x and a target specified at 
    construction.
    
    Both target and input need to be strictly positive.
    
    Arguments:
        target (torch.Tensor): Tensor to match. Broadcastable with input
    """
    def __init__(self, target):
        self.target = target

    @varargin
    def __call__(self, x):
        return -F.poisson_nll_loss(x, self.target, log_input=False)



import itertools
import numpy as np
import torch
from torch import nn

from .utils.img_similarity_metrics import cosine_similarity


def get_mask_of_point(
    point,
    grid_points_per_dim,
    num_invariances,
    neighbor_size,
    with_periodic_invariances,
    with_round_neighbor=False,
):
    assert num_invariances == 1 or num_invariances == 2
    ns = int(neighbor_size * grid_points_per_dim)
    if num_invariances == 1:
        assert isinstance(with_periodic_invariances, bool)
        mask = -torch.ones([grid_points_per_dim] * num_invariances)
        x = point[0]
        if x < ns:
            mask[: x + ns + 1] = 1
            if with_periodic_invariances:
                mask[x - ns :] = 1
        elif x + ns >= grid_points_per_dim:
            mask[x - ns :] = 1
            if with_periodic_invariances:
                mask[: ns - grid_points_per_dim + x + 1] = 1
        else:
            mask[x - ns : x + ns + 1] = 1
        mask[x] = 0
        return mask
    if num_invariances == 2:
        point = list(point)
        if isinstance(with_periodic_invariances, bool):
            with_periodic_invariances = [with_periodic_invariances] * 2
        if isinstance(with_periodic_invariances, list):
            assert len(with_periodic_invariances) == 2
        mask_n_points = grid_points_per_dim + 2 * ns
        mask_size = [grid_points_per_dim] * 2
        for dim, periodicity in enumerate(with_periodic_invariances):
            if not periodicity:
                mask_size[dim] = mask_n_points
                point[dim] = point[dim] + ns
        mask = -torch.ones(mask_size)
        start = [int(s / 2) for s in mask_size]
        mask[start[0] - ns : start[0] + ns + 1, start[1] - ns : start[1] + ns + 1] = 1
        mask[start[0], start[1]] = 0
        if with_round_neighbor:
            pos_idxs = np.argwhere(mask > 0)
            pos_idxs = np.array(pos_idxs).T

            for idx in pos_idxs:
                idx = tuple(idx)
                dist_x = idx[0] - start[0]
                dist_y = idx[1] - start[1]
                dist = np.sqrt(dist_x**2 + dist_y**2)
                if dist > ns:
                    mask[idx] = -1
        translation = [p - s for p, s in zip(point, start)]

        mask = np.roll(mask, shift=tuple(translation), axis=(0, 1))

        if not with_periodic_invariances[0]:
            mask = mask[ns : ns + grid_points_per_dim, :]
        if not with_periodic_invariances[1]:
            mask = mask[:, ns : ns + grid_points_per_dim]
        return mask


def GetPosNegMask(
    num_invariances,
    grid_points_per_dim,
    neighbor_size,
    with_periodic_invariances,
    with_round_neighbor,
):
    masks = np.ones([grid_points_per_dim] * 2 * num_invariances)
    masks = np.ones([grid_points_per_dim] * 2 * num_invariances)
    for point in itertools.product(
        np.arange(grid_points_per_dim), repeat=num_invariances
    ):
        masks[tuple(point)] = get_mask_of_point(
            point,
            grid_points_per_dim,
            num_invariances,
            neighbor_size,
            with_periodic_invariances,
            with_round_neighbor,
        )
    return masks


class SimCLROnGrid(nn.Module):
    def __init__(
        self,
        num_invariances,
        grid_points_per_dim,
        neighbor_size,
        temperature,
        with_periodic_invariances,
        with_round_neighbor=False,
        **args,
    ):
        super().__init__()
        self.num_invariances = num_invariances
        self.points_per_dim = grid_points_per_dim
        self.neighbor_size = neighbor_size
        self.temperature = temperature
        self.with_round_neighbor = with_round_neighbor
        self.eps = 1e-6
        self.neighbor_mask = torch.tensor(
            GetPosNegMask(
                num_invariances,
                grid_points_per_dim,
                neighbor_size,
                with_periodic_invariances,
                with_round_neighbor,
            ).astype(np.float32)
        )
        self.register_buffer(
            "flat_neighbor_mask",
            self.neighbor_mask.reshape(
                self.points_per_dim**self.num_invariances,
                self.points_per_dim**self.num_invariances,
            ),
        )
        self.register_buffer(
            "flat_neighbor_mask_pos",
            torch.where(self.flat_neighbor_mask > 0, self.flat_neighbor_mask, 0.0),
        )
        self.register_buffer(
            "flat_neighbor_mask_neg",
            torch.where(self.flat_neighbor_mask < 0, self.flat_neighbor_mask, 0.0) * -1,
        )

    def reg_term(self, images):
        images = images.flatten(
            1, -1
        )  # [N, c, h, w] -> [N, c * h * w]  N=number of points in the grid
        similarity = torch.exp(cosine_similarity(images) / self.temperature)
        positive_sim = (similarity * self.flat_neighbor_mask_pos).sum(dim=-1)
        positive_n = self.flat_neighbor_mask_pos.sum(dim=-1)
        # num = positive_sim / positive_n
        num = torch.where(positive_sim == 0, 0.0, positive_sim / positive_n) + self.eps

        negative_sim = (similarity * self.flat_neighbor_mask_neg).sum(dim=-1)
        negative_n = self.flat_neighbor_mask_neg.sum(dim=-1)
        # den = negative_sim / negative_n
        den = torch.where(negative_sim == 0, 0.0, negative_sim / negative_n) + self.eps
        reg_term = torch.log(num / den).mean()
        return reg_term * self.temperature / 2

    def pos_term(self, images):
        images = images.flatten(1, -1)
        similarity = torch.exp(cosine_similarity(images) / self.temperature)
        num = (similarity * self.flat_neighbor_mask_pos).sum(
            dim=-1
        ) / self.flat_neighbor_mask_pos.sum(dim=-1)
        return torch.log(num).mean() * self.temperature / 2

    def neg_term(self, images):
        images = images.flatten(1, -1)
        similarity = torch.exp(cosine_similarity(images) / self.temperature)
        den = -(-similarity * self.flat_neighbor_mask_neg).sum(
            dim=-1
        ) / self.flat_neighbor_mask_neg.sum(dim=-1)
        return -torch.log(den).mean() * self.temperature / 2



from collections import OrderedDict
import torch
import torch.nn as nn
import numpy as np
from .coordinate_transforms import CoordinateTransform


class INRTemplates(nn.Module):
    def __init__(
        self,
        img_res,
        num_neurons=0,
        num_templates=1,
        out_channels=1,
        aux_dim=1,
        # CPPN-related arguments
        widths=[15] * 8,
        periodic_invariance=False,
        positional_encoding_dim=None,
        positional_encoding_projection_scale=1.0,
        aux_positional_encoding_dim=None,
        aux_positional_encoding_projection_scale=1.0,
        nonlinearity=nn.Tanh,
        final_nonlinearity=nn.Sigmoid,
        bias=False,
        batchnorm=False,
        weights_scale=1.0,
        # Coordinate-related arguments
        # coordinate_transform_clamp_boundaries=True,
        # only_affine_coordinate_transformation=False,
        # stochastic_coordinate_transformation=False,
        # allow_scale_coordinate_transformation=True,
        # allow_shear_coordinate_transformation=True,
        # uniform_scale_coordinate_transformation=False,
        # init_noise_scale_coordinate_transformation=0.1,
        jitter_coords=False,
        jitter_coords_scale=1.0,
    ):
        super().__init__()
        self.setup_attributes(
            num_neurons,
            img_res,
            num_templates,
            out_channels,
            aux_dim,
            batchnorm,
            weights_scale,
            periodic_invariance,
            positional_encoding_dim,
            positional_encoding_projection_scale,
            aux_positional_encoding_dim,
            aux_positional_encoding_projection_scale,
            # coordinate_transform_clamp_boundaries,
            jitter_coords,
            jitter_coords_scale,
        )
        self.build_network(widths, nonlinearity, final_nonlinearity, bias)
        self.initialize_buffers()
        # if num_neurons > 0:
        #     self.initialize_coordinate_transform(
        #         only_affine_coordinate_transformation,
        #         stochastic_coordinate_transformation,
        #         init_noise_scale_coordinate_transformation,
        #         allow_scale=allow_scale_coordinate_transformation,
        #         allow_shear=allow_shear_coordinate_transformation,
        #         uniform_scale=uniform_scale_coordinate_transformation,
        #     )
        self.apply(self.weights_init)

    def forward(
        self, img_res=None, aux=None, return_template=False, centralize_template=False
    ):
        return_template = True if self.num_neurons == 0 else return_template
        device = next(self.parameters()).device
        coords = self.get_coordinates(img_res, device)
        if img_res is None:
            img_res = self.img_res

        coords_transformed = self.apply_coordinate_transform(
            coords, return_template, centralize_template
        )
        aux_transformed = self.transform_auxiliary_input(aux, device)
        coords_transformed_pe, aux_transformed_pe = self.apply_positional_encoding(
            coords_transformed, aux_transformed
        )

        # self.template_input.shape -> T
        # aux_transformed_pe.shape -> N x 1 (or anything depending on the positional encoding)
        # coords_transformed_pe.shape -> D x HW x 2 (or anything depending on the positional encoding)
        # T: num_templates, N: num_images, D: num_neurons, HW: flattened spatial dims

        inputs = combine_multiple_tensors(
            self.template_input, aux_transformed_pe, coords_transformed_pe
        )
        T, N, D, HW = tensors_coordinate_dims(
            self.template_input, aux_transformed_pe, coords_transformed_pe
        )

        assert HW == np.prod(img_res)
        out = (
            self.func(inputs)
            .reshape(T, N, D, *img_res, self.out_channels)
            .permute(0, 2, 1, 5, 3, 4)
        )

        output_aux_dim_unpacked = [round(len(aux) ** (1 / self.aux_dim))] * self.aux_dim

        if return_template:
            return (
                out.squeeze(0)
                .squeeze(0)
                .reshape(*output_aux_dim_unpacked, self.out_channels, *img_res)
            )
        else:
            return out.squeeze(0).reshape(
                D, *output_aux_dim_unpacked, self.out_channels, *img_res
            )

    def setup_attributes(
        self,
        num_neurons,
        img_res,
        num_templates,
        out_channels,
        aux_dim,
        batchnorm,
        weights_scale,
        periodic_invariance,
        positional_encoding_dim,
        positional_encoding_projection_scale,
        aux_positional_encoding_dim,
        aux_positional_encoding_projection_scale,
        # coordinate_transform_clamp_boundaries,
        jitter_coords,
        jitter_coords_scale,
    ):
        self.img_res = img_res
        self.out_channels = out_channels
        self.batchnorm = batchnorm
        self.num_neurons = num_neurons
        self.num_templates = num_templates
        self.aux_dim = aux_dim
        self.template_dim = 1
        self.weights_scale = weights_scale
        self.periodic_invariance = periodic_invariance
        self.positional_encoding_dim = positional_encoding_dim
        self.positional_encoding_projection_scale = positional_encoding_projection_scale
        self.aux_positional_encoding_dim = aux_positional_encoding_dim
        self.aux_positional_encoding_projection_scale = (
            aux_positional_encoding_projection_scale
        )
        # self.coordinate_transform_clamp_boundaries = (
        #     coordinate_transform_clamp_boundaries
        # )
        self.jitter_coords = jitter_coords
        self.jitter_coords_scale = jitter_coords_scale
        self.calculate_input_dimensions()

    def calculate_input_dimensions(self):
        if self.positional_encoding_dim is not None:
            self.in_dim = 2 * self.positional_encoding_dim + self.template_dim
        else:
            self.in_dim = 2 + self.template_dim

        if self.aux_positional_encoding_dim is not None:
            self.in_dim += self.aux_positional_encoding_dim * 2
        else:
            if self.periodic_invariance:
                self.in_dim += self.aux_dim * 2
            else:
                self.in_dim += self.aux_dim

    def build_network(self, widths, nonlinearity, final_nonlinearity, bias):
        layers = len(widths)
        in_widths = [self.in_dim] + widths[:-1]
        elements = []
        for i in range(layers):
            elements.append(
                (f"layer{i}", nn.Linear(in_widths[i], widths[i], bias=bias))
            )
            if self.batchnorm:
                elements.append((f"batchnorm{i}", nn.BatchNorm1d(widths[i])))
            if nonlinearity is not None:
                elements.append((f"nonlinearity{i}", nonlinearity()))

        elements.append(
            (f"layer{layers}", nn.Linear(widths[-1], self.out_channels, bias=bias))
        )
        elements.append((f"nonlinearity{layers}", final_nonlinearity()))

        self.func = nn.Sequential(OrderedDict(elements))

    def initialize_buffers(self):
        if self.positional_encoding_dim is not None:
            self.register_buffer(
                "B",
                torch.randn(2, self.positional_encoding_dim)
                * self.positional_encoding_projection_scale,
            )
        if self.aux_positional_encoding_dim is not None:
            auxB_shape = (
                (2 * self.aux_dim, self.aux_positional_encoding_dim)
                if self.periodic_invariance
                else (1 * self.aux_dim, self.aux_positional_encoding_dim)
            )
            self.register_buffer(
                "auxB",
                torch.randn(auxB_shape) * self.aux_positional_encoding_projection_scale,
            )

        coords = torch.meshgrid(
            [torch.linspace(-1, 1, linres) for linres in self.img_res], indexing="ij"
        )
        coords = torch.stack(coords, dim=-1)
        self.register_buffer("coords", coords)

        template_input = (
            torch.linspace(-1, 1, self.num_templates)
            if self.num_templates > 1
            else torch.zeros(1)
        )
        self.register_buffer("template_input", template_input)

    def weights_init(self, m):
        if isinstance(m, nn.Linear):
            m.weight.data.normal_(0.0, self.weights_scale)
            if m.bias is not None:
                m.bias.data.fill_(0)

    def initialize_coordinate_transform(
        self,
        num_neurons,
        only_affine,
        stochastic,
        init_noise_scale,
        allow_scale,
        allow_shear,
        uniform_scale,
        clamp_boundaries,
    ):
        device = next(self.parameters()).device
        self.num_neurons = num_neurons
        self.coordinate_transform = CoordinateTransform(
            self.num_neurons,
            only_affine=only_affine,
            stochastic=stochastic,
            init_noise_scale=init_noise_scale,
            allow_scale=allow_scale,
            allow_shear=allow_shear,
            uniform_scale=uniform_scale,
        ).to(device)
        self.coordinate_transform_clamp_boundaries = clamp_boundaries

    def reset_coordinate_transform(self, num_neurons=None):
        device = next(self.parameters()).device #TODO: use infer_device
        if num_neurons is not None:
            self.num_neurons = num_neurons
        else:
            num_neurons = self.coordinate_transform.n_dim
        only_affine = self.coordinate_transform.only_affine
        stochastic = self.coordinate_transform.stochastic
        init_noise_scale = self.coordinate_transform.init_noise_scale
        allow_scale = self.coordinate_transform.allow_scale
        allow_shear = self.coordinate_transform.allow_shear
        uniform_scale = self.coordinate_transform.uniform_scale

        self.coordinate_transform = CoordinateTransform(
            self.num_neurons,
            only_affine=only_affine,
            stochastic=stochastic,
            init_noise_scale=init_noise_scale,
            allow_scale=allow_scale,
            allow_shear=allow_shear,
            uniform_scale=uniform_scale,
        ).to(device)

    def get_coords_shift_to_center(self, template_rf_center, img_res=None):
        template_rf_center = template_rf_center.flip(dims=[-1])
        device = next(self.parameters()).device
        img_res = img_res if img_res is not None else self.img_res
        img_center_coord = torch.tensor(img_res).to(device) / 2
        coords_shift = (
            template_rf_center.to(device) - img_center_coord
        ) / img_center_coord
        return coords_shift

    def get_coords_shift_from_center(self, template_rf_center, img_res=None):
        shift_to_center = self.get_coords_shift_to_center(
            template_rf_center, img_res=img_res
        )
        return -1 * shift_to_center

    def get_template_center_in_coords_space(self, template_rf_center, img_res=None):
        template_rf_center = template_rf_center.flip(dims=[-1])
        device = next(self.parameters()).device
        img_res = img_res if img_res is not None else self.img_res
        img_max_coord = torch.tensor(img_res).to(device)
        template_center = template_rf_center.to(device) / img_max_coord * 2 - 1
        return template_center

    def register_coords_shifts(self, template_rf_center, target_rf_centers):
        self.register_buffer(
            "coords_shift_to_center",
            self.get_coords_shift_to_center(template_rf_center),
        )
        self.register_buffer(
            "template_center_in_coords_space",
            self.get_template_center_in_coords_space(template_rf_center),
        )
        self.register_buffer(
            "coords_shift_from_center",
            self.get_coords_shift_from_center(target_rf_centers),
        )

    def get_coordinates(self, img_res, device):
        if img_res is None:
            coords = self.coords
            img_res = self.img_res
        else:
            coords = torch.stack(
                torch.meshgrid(
                    [torch.linspace(-1, 1, shape, device=device) for shape in img_res],
                    indexing="ij",
                ),
                dim=-1,
            )
        if self.jitter_coords and self.training:  # only jitter during training
            coords_noise = (torch.rand_like(coords) - 0.5) * torch.tensor(
                [[[2 / img_res[0], 2 / img_res[1]]]]
            ).to(device)
            coords = coords + coords_noise * self.jitter_coords_scale
        return coords

    def apply_coordinate_transform(self, coords, return_template, centralize_template):
        """
        Three important steps here:
        1. Get grid points relative to the center of rotation (template's center) -> this is necessary to apply a local transformation (other than translation)
        2. Apply the local transformation (e.g. rotation) per neuron
        3. Locally transformed coordinates are then shifted to the target neurons' RF location
        """
        coords_shift_to_center = getattr(
            self,
            "coords_shift_to_center",
            torch.zeros(self.num_templates, 2).to(coords.device),
        )
        coords_shifted_to_center = coords.flatten(end_dim=-2) + coords_shift_to_center
        if return_template:
            return (
                coords_shifted_to_center.unsqueeze(0)
                if centralize_template
                else coords.flatten(end_dim=-2).unsqueeze(0)
            )
        else:
            template_center = getattr(
                self,
                "template_center_in_coords_space",
                torch.zeros(self.num_templates, 2).to(coords.device),
            )
            coords_relative_to_template_center = (
                coords.flatten(end_dim=-2) - template_center
            )  # center the grid on the location of interest
            coords_transformed = (
                self.coordinate_transform(coords_relative_to_template_center)
                + template_center
            )  # transform back to original location

            # Apply the final translation to brind the transformed grid points at the location of neurons' RF
            with torch.no_grad():
                coords_shift_from_center = getattr(
                    self,
                    "coords_shift_from_center",
                    torch.zeros(self.num_neurons, 2).to(coords.device),
                )
                coords_shift_from_center_adjusted = coords_shift_from_center.unsqueeze(
                    1
                ) + coords_shift_to_center.unsqueeze(0)
                coords_shift_from_center_transformed = self.coordinate_transform(
                    coords_shift_from_center_adjusted
                )

            coords_transformed_shifted_to_target = (
                coords_transformed + coords_shift_from_center_transformed
            )
            if self.coordinate_transform_clamp_boundaries:
                coords_transformed_shifted_to_target.data.clamp_(-1, 1)
            return coords_transformed_shifted_to_target

    def transform_auxiliary_input(self, aux, device):
        if aux is None:
            aux = torch.tensor([[0.0]]).to(device)
        if len(aux.shape) == 1:
            aux = aux.unsqueeze(1)
        # aux = combine_multiple_tensors(*aux.T)
        if self.periodic_invariance:
            aux_transformed = torch.cat(
                [torch.sin(aux), torch.cos(aux)],
                dim=1,
            )
        else:
            aux_transformed = (
                aux / np.pi - 1
            )  # the assumption is that the aux is between 0 and 2*pi
        return aux_transformed

    def apply_positional_encoding(self, coords_transformed, aux_transformed):
        # Apply positional encoding to coordinates
        if self.positional_encoding_dim is not None:
            coords_transformed_pe = torch.cat(
                [
                    torch.sin(coords_transformed @ self.B),
                    torch.cos(coords_transformed @ self.B),
                ],
                dim=-1,
            )
        else:
            coords_transformed_pe = coords_transformed

        # Apply positional encoding to auxiliary input
        if self.aux_positional_encoding_dim is not None:
            aux_transformed_pe = torch.cat(
                [
                    torch.sin(aux_transformed @ self.auxB),
                    torch.cos(aux_transformed @ self.auxB),
                ],
                dim=-1,
            )
        else:
            aux_transformed_pe = aux_transformed

        return coords_transformed_pe, aux_transformed_pe


def tensor_coordinate_dims(tensor):
    dims = len(tensor.shape)
    if dims == 1:
        return [tensor.shape[0]]
    else:
        return list(tensor.shape[:-1])


def tensors_coordinate_dims(*tensors):
    return [dim for tensor in tensors for dim in tensor_coordinate_dims(tensor)]


def reshape_tensor(tensor):
    dims = len(tensor.shape)
    if dims == 1:
        return tensor.view(-1, 1)
    else:
        return tensor.reshape(-1, tensor.shape[-1])


def cartesian_product(A, B):
    m, n = A.shape
    p, q = B.shape
    A = A[:, None, :]
    B = B[None, :, :]
    A_comb = A.repeat(1, p, 1)
    B_comb = B.repeat(m, 1, 1)
    C = torch.cat((A_comb, B_comb), dim=-1)
    return C.reshape(m * p, n + q)


def combine_multiple_tensors(*tensors):
    reshaped_tensors = [reshape_tensor(tensor) for tensor in tensors]
    result = reshaped_tensors[0]
    for T in reshaped_tensors[1:]:
        result = cartesian_product(result, T)
    return result



from typing import Any, Mapping, Optional, Tuple

import torch
from torch.nn import functional as F
from torch.nn import init
from torch.nn.modules import ELU, BatchNorm2d, Conv2d, Module, Sequential
from torch.nn.parameter import Parameter

from .base import Readout, Reduction


class AttentionReadout(Readout):
    def __init__(
        self,
        in_shape: Tuple[int, int, int],
        outdims: int,
        bias: bool,
        init_noise: float = 1e-3,
        attention_kernel: int = 1,
        attention_layers: int = 1,
        mean_activity: Optional[torch.Tensor] = None,
        feature_reg_weight: float = 1.0,
        gamma_readout: Optional[float] = None,  # deprecated, use feature_reg_weight instead
        **kwargs: Any,
    ) -> None:
        super().__init__()  # type: ignore[no-untyped-call]
        self.in_shape = in_shape
        self.outdims = outdims
        self.feature_reg_weight = self.resolve_deprecated_gamma_readout(feature_reg_weight, gamma_readout)
        self.mean_activity = mean_activity
        c, w, h = in_shape
        self.features = Parameter(torch.Tensor(self.outdims, c))

        attention = Sequential()
        for i in range(attention_layers - 1):
            attention.add_module(
                f"conv{i}",
                Conv2d(c, c, attention_kernel, padding=attention_kernel > 1),
            )
            attention.add_module(f"norm{i}", BatchNorm2d(c))  # type: ignore[no-untyped-call]
            attention.add_module(f"nonlin{i}", ELU())
        else:
            attention.add_module(
                f"conv{attention_layers}",
                Conv2d(c, outdims, attention_kernel, padding=attention_kernel > 1),
            )
        self.attention = attention

        self.init_noise = init_noise
        if bias:
            bias_param = Parameter(torch.Tensor(self.outdims))
            self.register_parameter("bias", bias_param)
        else:
            self.register_parameter("bias", None)
        self.initialize(mean_activity)

    @staticmethod
    def init_conv(m: Module) -> None:
        if isinstance(m, Conv2d):
            init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                m.bias.data.fill_(0)

    def initialize_attention(self) -> None:
        self.apply(self.init_conv)

    def initialize(self, mean_activity: Optional[torch.Tensor] = None) -> None:  # type: ignore[override]
        if mean_activity is None:
            mean_activity = self.mean_activity
        self.features.data.normal_(0, self.init_noise)
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)
        self.initialize_attention()

    def feature_l1(self, reduction: Reduction = "sum", average: Optional[bool] = None) -> torch.Tensor:
        return self.apply_reduction(self.features.abs(), reduction=reduction, average=average)

    def regularizer(self, reduction: Reduction = "sum", average: Optional[bool] = None) -> torch.Tensor:
        return self.feature_l1(reduction=reduction, average=average) * self.feature_reg_weight

    def forward(self, x: torch.Tensor, shift: Optional[Any] = None) -> torch.Tensor:
        attention = self.attention(x)
        b, c, w, h = attention.shape
        attention = F.softmax(attention.view(b, c, -1), dim=-1).view(b, c, w, h)
        y: torch.Tensor = torch.einsum("bnwh,bcwh->bcn", attention, x)  # type: ignore[attr-defined]
        y = torch.einsum("bcn,nc->bn", y, self.features)  # type: ignore[attr-defined]
        if self.bias is not None:
            y = y + self.bias
        return y

    def __repr__(self) -> str:
        return self.__class__.__name__ + " (" + "{} x {} x {}".format(*self.in_shape) + " -> " + str(self.outdims) + ")"



from collections import namedtuple

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import numpy as np
import torch
from skimage.transform import rescale


class Invertible:
    def inv(self, y):
        raise NotImplemented("Subclasses of Invertible must implement an inv method")


class DataTransform:
    def __repr__(self):
        return self.__class__.__name__

    def id_transform(self, id_map):
        """
        Given a dictionary mapping from data group name to
        a numpy array of appropriate size containing identity information
        for that particular group name, the transform is expected to return
        the expected identity modification (e.g. dropping or duplication of entries, reordering, etc)
        """
        return id_map

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class MovieTransform(DataTransform):
    """
    Abstract class to certify that the transform is valid for sequence like (e.g. movie) datasets.
    """


class StaticTransform(DataTransform):
    """
    Abstract class to certify that the transform is valid for non-sequential (e.g. image) datasets.
    """


class Subsequence(MovieTransform):
    def __init__(self, frames, channel_first=("inputs",), offset=-1):
        """
        Given a sequential (movie like) data, subselect a consequent `frames` counts of frames, starting with
        `offset` frames skipped. If `offset`< 0, then the subsequence is taken with a random (but valid) offset each iteration.

        Args:
            frames (int): Length of subsequence to be selected from each sample
            channel_first (tuple, optional): A list of data key names where the channel (and thus not time) dimension occurs on the first dimension (dim=0). Otherwise, it's assumed
            that the time dimesion occurs on the first dimension. Defaults to ("inputs",).
            offset (int, optional): Offset to start the subsequence from. Defaults to -1, corresponding to random but valid offset at each iteration.
        """
        self.frames = frames
        # for special group, the time slicing is applied on
        self.channel_first = channel_first
        self.offset = offset

    def __call__(self, x):
        first_group = x._fields[0]

        # get the length of the full sequence. Note that which axis to pick depends on if it's
        # channel first (i.e. time second) group
        t = getattr(x, first_group).shape[int(first_group in self.channel_first)]

        if self.offset < 0:
            i = np.random.randint(0, t - self.frames)
        else:
            i = self.offset
        return x.__class__(
            **{
                k: getattr(x, k)[:, i : i + self.frames, ...]
                if k in self.channel_first
                else getattr(x, k)[i : i + self.frames, ...]
                for k in x._fields
            }
        )

    def id_transform(self, id_map):
        # until a better solution is reached, skipping this
        return id_map

        new_map = {}
        first_group = list(id_map.keys())[0]
        v_fg = id_map[first_group]
        t = v_fg.shape[int(first_group in self.channel_first)]
        i = np.random.randint(0, t - self.frames)
        for k, v in id_map.items():
            if k in self.channel_first:
                new_map[k] = v[:, i : i + self.frames]
            else:
                new_map[k] = v[i : i + self.frames]

        return new_map

    def __repr__(self):
        return self.__class__.__name__ + "({})".format(self.frames)


class Delay(MovieTransform):
    """
    Delay the specified target gorups by delay frames. In other words,
    given non-delayed group g(t) and delayed group d(t),
    g(T:N-delay) will be returned with d(T+delay:N) where `delay` is the specified amount
    of delay, and N is the last frame in the dataset.
    """

    def __init__(self, delay, delay_groups=("responses",), channel_first=("inputs",)):
        self.delay = delay
        self.delay_groups = delay_groups
        self.channel_first = channel_first

    def __call__(self, x):
        first_group = x._fields[0]
        t = getattr(x, first_group).shape[int(first_group in self.channel_first)]
        assert t > self.delay, "The sequence length {} has to be longer than the delay {}".format(t, self.delay)
        key_entry = {}
        for k in x._fields:
            if k in self.delay_groups:
                start, stop = self.delay, t
            else:
                start = 0
                stop = t - self.delay

            key_entry[k] = getattr(x, k)[:, start:stop] if k in self.channel_first else getattr(x, k)[start:stop]

        return x.__class__(**key_entry)

    def id_transform(self, id_map):
        # until a better solution is reached, skipping this
        return id_map

        new_map = {}
        first_group = list(id_map.keys())[0]
        v_fg = id_map[first_group]
        t = v_fg.shape[int(first_group in self.channel_first)]
        assert t > self.delay, "The sequence length {} has to be longer than the delay {}".format(t, self.delay)

        for k, v in id_map.items():
            if k in self.delay_groups:
                start, end = self.delay, t
            else:
                start, end = 0, t - self.delay

            new_map[k] = v[:, start:end] if k in self.channel_first else v[start:end]

        return new_map

    def __repr__(self):
        return self.__class__.__name__ + "({} on {})".format(self.delay, self.delay_groups)


class Stack(MovieTransform):
    def __init__(
        self,
        target="inputs",
        sources=("eye_pos", "behavior"),
        concat_axis=0,
        transpose=True,
    ):
        """
        Stack source data elements into the target data elements. In stacking, the source data elements are
        left aligned with the target, and it's dimensions expanded as necessary before stacking along the
        specified existing axis.

        Examples:
        If target is an array of shape [1, 150, 36, 64], and one of the sources is of shape [3, 150], then
        the source data is first expanded into [3, 150, 1, 1], followed by tiling to achieve [3, 150, 36, 64],
        and this is finally stacked with the target to yield a new output of shape [4, 150, 36, 64], where the
        output[0, ...] is the original target, and output[1:4] is the expanded source data. If `tranpose=True`,
        the source is transposed first before performing dimension alignment and expansions.

        Args:
            target (str, optional): Data key for the target to be modified with stacking. Defaults to "inputs".
            sources (str or tuple, optional): A single source or atuple of sources to be stacked into the target.
                Defaults to ("eye_pos", "behavior").
            concat_axis (int, optional): Axis along which sources are concatenated into the target. Defaults to 0.
            transpose (bool, optional): Whether to transpose the sources first. Defaults to True.
        """
        self.target = target
        if isinstance(sources, str):
            sources = (sources,)
        self.sources = sources
        self.concat_axis = concat_axis
        self.transpose = transpose

    def __call__(self, x):
        x_dict = x._asdict()
        target = x_dict[self.target]
        groups = [target]
        for source in [x_dict[s] for s in self.sources]:
            if self.transpose:
                source = source.T
            n_target = len(target.shape)
            n_source = len(source.shape)
            dims = list(range(-n_target + n_source, 0))
            groups.append(np.ones((1,) * n_source + target.shape[n_source:]) * np.expand_dims(source, axis=dims))
        x_dict[self.target] = np.concatenate(groups, axis=self.concat_axis)
        return x.__class__(**x_dict)

    def id_transform(self, id_map):
        # until a better solution is reached, skipping this
        return id_map

    def __repr__(self):
        items = ", ".join(s + ".T" if self.transpose else s for s in self.sources)
        return self.__class__.__name__ + "(stack [{}] on {} along axis={})".format(items, self.target, self.concat_axis)


class Subsample(MovieTransform, StaticTransform):
    def __init__(self, idx, target_group="responses", target_index=None):
        """
        Subselects samples for data_key specified by `target_group`. By default, the subselection is performed on
        the last index of the tensor, but this behavior may be modified by passing in `target_index` a dictionary
        mapping the position of the index to the name of the data_key.

        Args:
            idx (numpy index specifier): Indices to be selected. Must be a valid NumPy index specification (e.g. list of indicies, boolean array, etc.)
            target_group (string or iterable of strings): Specifies the taget data key to perform subselection on. If given a string, it is assumed as the direct name of data_key
                Otherwise, it is assumed to be an iterable over string values of all data_keys
            target_index (optional, dict): If provided a dictionary, the key is asssumed to be the name of data_key and value the index position to perform subselection on. If not provided, index position of -1 (last position) is used.
        """

        self.idx = idx
        if isinstance(target_group, str):
            self.target_groups = (target_group,)
        else:
            self.target_groups = target_group

        if target_index is None:
            target_index = {k: -1 for k in self.target_groups}
        elif isinstance(target_index, int):
            target_index = {k: target_index for k in self.target_groups}
        self.target_index = target_index

        assert np.ndim(self.idx) == 1, "Dimensionality of index array has to be 1"

    def __call__(self, x):
        return x.__class__(
            **{
                k: np.take(getattr(x, k), self.idx, self.target_index[k]) if k in self.target_groups else getattr(x, k)
                for k in x._fields
            }
        )

    def __repr__(self):
        return self.__class__.__name__ + "(n={})".format(len(self.idx))

    def id_transform(self, id_map):
        return {k: v[self.idx] if k in self.target_groups else v for k, v in id_map.items()}


class ToTensor(MovieTransform, StaticTransform, Invertible):
    def __init__(self, cuda=False):
        self.cuda = cuda

    def inv(self, y):
        return y.numpy()

    def __call__(self, x):
        return x.__class__(
            *[
                torch.from_numpy(elem.astype(np.float32)).cuda()
                if self.cuda
                else torch.from_numpy(elem.astype(np.float32))
                for elem in x
            ]
        )


class Identity(MovieTransform, StaticTransform, Invertible):
    def __call__(self, x):
        return x

    def inv(self, y):
        return y


class Rename(MovieTransform, StaticTransform, Invertible):
    def __init__(self, name_map):
        self.name_map = name_map
        self.rev_map = {v: k for k, v in name_map.items()}
        self.tuple_class = None
        self.origin_tuple_class = None

    def __call__(self, x):
        if self.tuple_class is None:
            if self.origin_tuple_class is None:
                self.origin_tuple_class = x.__class__
            renamed_fields = [self.name_map.get(f, f) for f in x._fields]
            self.tuple_class = namedtuple("RenamedDataPoint", renamed_fields)
        return self.tuple_class(*x)

    def inv(self, y):
        if self.origin_tuple_class is None:
            renamed_fields = [self.rev_map.get(f, f) for f in y._fields]
            self.origin_tuple_class = namedtuple("OriginalDataPoint", renamed_fields)
        return self.origin_tuple_class(*y)

    def id_transform(self, id_map):
        return {self.name_map.get(k, k): v for k, v in id_map.items()}


class NeuroNormalizer(MovieTransform, StaticTransform, Invertible):
    """
    Note that this normalizer only works with datasets that provide specific attributes information
    of very specific formulation

    Normalizes a trial with fields: inputs, behavior, eye_position, and responses. The pair of
    behavior and eye_position can be missing. The following normalizations are applied:

    - inputs are scaled by the training std of the stats_source and centered on the mean of the movie
    - behavior is divided by the std if the std is greater than 1% of the mean std (to avoid division by 0)
    - eye_position is z-scored
    - reponses are divided by the per neuron std if the std is greater than
            1% of the mean std (to avoid division by 0)
    """

    def __init__(self, data, stats_source="all", exclude=None, inputs_mean=None, inputs_std=None):

        self.exclude = exclude or []

        in_name = "images" if "images" in data.statistics.keys() else "inputs"
        out_name = "responses" if "responses" in data.statistics.keys() else "targets"
        eye_name = "pupil_center" if "pupil_center" in data.data_keys else "eye_position"

        self._inputs_mean = data.statistics[in_name][stats_source]["mean"][()] if inputs_mean is None else inputs_mean
        self._inputs_std = data.statistics[in_name][stats_source]["std"][()] if inputs_mean is None else inputs_std

        s = np.array(data.statistics[out_name][stats_source]["std"])

        # TODO: consider other baselines
        threshold = 0.01 * s.mean()
        idx = s > threshold
        self._response_precision = np.ones_like(s) / threshold
        self._response_precision[idx] = 1 / s[idx]
        transforms, itransforms = {}, {}

        # -- inputs
        transforms[in_name] = lambda x: (x - self._inputs_mean) / self._inputs_std
        itransforms[in_name] = lambda x: x * self._inputs_std + self._inputs_mean

        # -- responses
        transforms[out_name] = lambda x: x * self._response_precision
        itransforms[out_name] = lambda x: x / self._response_precision

        # -- behavior
        transforms["behavior"] = lambda x: x

        # -- trial_idx
        trial_idx_mean = np.arange(data._len).mean()
        trial_idx_std = np.arange(data._len).std()
        transforms["trial_idx"] = lambda x: (x - trial_idx_mean) / trial_idx_std
        itransforms["trial_idx"] = lambda x: x * trial_idx_std + trial_idx_mean

        if eye_name in data.data_keys:
            self._eye_mean = np.array(data.statistics[eye_name][stats_source]["mean"])
            self._eye_std = np.array(data.statistics[eye_name][stats_source]["std"])
            transforms[eye_name] = lambda x: (x - self._eye_mean) / self._eye_std
            itransforms[eye_name] = lambda x: x * self._eye_std + self._eye_mean

        if "behavior" in data.data_keys:
            s = np.array(data.statistics["behavior"][stats_source]["std"])

            self._behavior_precision = 1 / s
            # -- behavior
            transforms["behavior"] = lambda x: x * self._behavior_precision
            itransforms["behavior"] = lambda x: x / self._behavior_precision

        self._transforms = transforms
        self._itransforms = itransforms

    def __call__(self, x):
        """
        Apply transformation
        """
        return x.__class__(
            **{k: (self._transforms[k](v) if k not in self.exclude else v) for k, v in zip(x._fields, x)}
        )

    def inv(self, x):
        return x.__class__(
            **{k: (self._itransforms[k](v) if k not in self.exclude else v) for k, v in zip(x._fields, x)}
        )

    def __repr__(self):
        return super().__repr__() + ("(not {})".format(", ".join(self.exclude)) if self.exclude is not None else "")


class AddBehaviorAsChannels(MovieTransform, StaticTransform, Invertible):
    """
    Given a StaticImage object that includes "images", "responses", and "behavior", it returns three variables:
        - input image concatinated with behavior as new channel(s)
        - responses
        - behavior
    """

    def __init__(self):
        self.transforms, self.itransforms = {}, {}
        self.transforms["images"] = lambda img, behavior: np.concatenate(
            (
                img,
                np.ones((1, *img.shape[-(len(img.shape) - 1) :]))
                * np.expand_dims(behavior, axis=((len(img.shape) - 2), (len(img.shape) - 1))),
            ),
            axis=len(img.shape) - 3,
        )
        self.transforms["responses"] = lambda x: x
        self.transforms["behavior"] = lambda x: x
        self.transforms["pupil_center"] = lambda x: x
        self.transforms["trial_idx"] = lambda x: x

    def __call__(self, x):

        key_vals = {k: v for k, v in zip(x._fields, x)}
        dd = {
            "images": self.transforms["images"](key_vals["images"], key_vals["behavior"]),
            "responses": self.transforms["responses"](key_vals["responses"]),
            "behavior": self.transforms["behavior"](key_vals["behavior"]),
        }
        if "pupil_center" in key_vals:
            dd["pupil_center"] = self.transforms["pupil_center"](key_vals["pupil_center"])
        if "trial_idx" in key_vals:
            dd["trial_idx"] = self.transforms["trial_idx"](key_vals["trial_idx"])
        return x.__class__(**dd)


class AddPupilCenterAsChannels(MovieTransform, StaticTransform, Invertible):
    """
    Given a StaticImage object that includes "images", "responses", and "pupil center", it returns three variables:
        - input image concatenated with eye position as new channel(s)
        - responses
        - behavior
        - pupil center
    """

    def __init__(self):
        self.transforms, self.itransforms = {}, {}
        self.transforms["images"] = lambda img, pupil_center: np.concatenate(
            (
                img,
                np.ones((1, *img.shape[-(len(img.shape) - 1) :]))
                * np.expand_dims(pupil_center, axis=((len(img.shape) - 2), (len(img.shape) - 1))),
            ),
            axis=len(img.shape) - 3,
        )
        self.transforms["responses"] = lambda x: x
        self.transforms["behavior"] = lambda x: x
        self.transforms["pupil_center"] = lambda x: x

    def __call__(self, x):

        key_vals = {k: v for k, v in zip(x._fields, x)}
        dd = {
            "images": self.transforms["images"](key_vals["images"], key_vals["pupil_center"]),
            "responses": self.transforms["responses"](key_vals["responses"]),
        }
        if "behavior" in key_vals:
            dd["behavior"] = self.transforms["behavior"](key_vals["behavior"])
        dd["pupil_center"] = self.transforms["pupil_center"](key_vals["pupil_center"])

        if "trial_idx" in key_vals:
            dd["trial_idx"] = self.transforms["trial_idx"](key_vals["trial_idx"])
        return x.__class__(**dd)


class SelectInputChannel(StaticTransform):
    """
    Given a StaticImage object that includes "images", it will select a particular input channel.
    """

    def __init__(self, grab_channel):
        self.grab_channel = grab_channel if isinstance(grab_channel, Iterable) else [grab_channel]

    def __call__(self, x):
        key_vals = {k: v for k, v in zip(x._fields, x)}
        img = key_vals["images"]
        key_vals["images"] = img[:, (self.grab_channel,)] if len(img.shape) == 4 else img[self.grab_channel, ...]
        return x.__class__(**key_vals)


class ScaleInputs(StaticTransform, Invertible):
    """
    Applies skimage.transform.rescale to the data_key "images".
    """

    def __init__(
        self,
        scale,
        mode="reflect",
        anti_aliasing=False,
        preserve_range=True,
        clip=True,
        in_name="images",
        channel_axis=0,
    ):

        self.scale = scale
        self.mode = mode
        self.anti_aliasing = anti_aliasing
        self.preserve_range = preserve_range
        self.clip = clip
        self.in_name = in_name
        self.channel_axis = channel_axis

    def __call__(self, x):
        key_vals = {k: v for k, v in zip(x._fields, x)}
        img = key_vals[self.in_name]
        key_vals[self.in_name] = rescale(
            img,
            scale=self.scale,
            mode=self.mode,
            anti_aliasing=self.anti_aliasing,
            clip=self.clip,
            preserve_range=self.preserve_range,
            channel_axis=self.channel_axis,
        )
        return x.__class__(**key_vals)



import numpy as np
import torch
from torch import nn
import matplotlib.pyplot as plt
from torch.nn import functional as F


class ArbitraryNeuronModel(nn.Module):
    def __init__(self, filters, eps=1e-6):
        super().__init__()
        self.register_buffer("filters", torch.from_numpy(filters.astype(np.float32)))
        self.eps = eps

    def forward(self, x):
        out = torch.einsum("bchw,fhw->bf", x, self.filters)
        max_values, _ = out.max(dim=1)
        return (F.elu(max_values) + 1) / 2 + self.eps


class ArbitraryMultiNeuronModel(nn.Module):
    def __init__(self, neuron_models_list):
        super().__init__()
        self.neuron_models_list = nn.ModuleList(neuron_models_list)

    def forward(self, x):
        preds = []
        for model in self.neuron_models_list:
            preds.append(model(x))
        return torch.stack(preds, dim=1)


class ComplexCell(nn.Module):
    def __init__(self, even_filter_np, odd_filter_np):
        super(ComplexCell, self).__init__()
        self.register_buffer(
            "even_filter", torch.tensor(even_filter_np, dtype=torch.float32)
        )
        self.register_buffer(
            "odd_filter", torch.tensor(odd_filter_np, dtype=torch.float32)
        )

    def forward(self, x):
        even_response = torch.sum(x * self.even_filter, dim=[2, 3])
        odd_response = torch.sum(x * self.odd_filter, dim=[2, 3])
        energy = even_response**2 + odd_response**2
        return torch.sqrt(energy)[:, 0]


def sigmoid(x, temp=1.0):
    return 1 / (1 + np.exp(-x / temp))


def transform_pattern(xx, yy, transformation_mat, center_of_transformation):
    p_x, p_y = center_of_transformation
    xx_r = xx - p_x
    yy_r = yy - p_y
    xy_r = np.stack((xx_r, yy_r)).reshape(2, -1)
    xy_rt = transformation_mat @ xy_r
    xx_rt, yy_rt = xy_rt.reshape(2, *xx.shape)
    xx, yy = xx_rt + p_x, yy_rt + p_y

    return xx, yy


def translate_pattern(xx, yy, shift_from_center):
    dx, dy = shift_from_center
    xx, yy = xx - dx, yy - dy
    return xx, yy


def generate_gabor_filter(
    theta,
    width_x,
    width_y,
    frequency,
    phase,
    center,
    grid_size,
    transformation_matrix=None,
):
    center_of_rotation = center
    shift_from_center = center

    x = np.linspace(-1, 1, grid_size[0])
    y = np.linspace(-1, 1, grid_size[1])
    xx, yy = np.meshgrid(x, y)

    # Apply Translation
    xx, yy = translate_pattern(xx, yy, shift_from_center)

    # Apply local rotation
    dx, dy = shift_from_center
    p_x, p_y = center_of_rotation
    new_center_of_rotation = (p_x - dx, p_y - dy)
    rotation_matrix = np.array(
        [[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]]
    )
    xx, yy = transform_pattern(xx, yy, rotation_matrix, new_center_of_rotation)

    # Apply goabl transformation
    if transformation_matrix is not None:
        dx, dy = shift_from_center
        p_x, p_y = 0, 0  # center of global transformation is the center
        new_center_of_transformation = (p_x - dx, p_y - dy)
        new_center_of_transformation_rotated = (
            rotation_matrix @ new_center_of_transformation
        )
        xx, yy = transform_pattern(
            xx, yy, transformation_matrix, new_center_of_transformation_rotated
        )

    gaussian_maks = np.exp(-0.5 * (xx**2 / width_x**2 + yy**2 / width_y**2))
    spatial_sinudoid = np.cos(2 * np.pi * frequency * xx + phase)
    gabor_filter = gaussian_maks * spatial_sinudoid
    return gabor_filter


def generate_gabor_filter_span_as_grid(center, grid_size, transformation_matrix=None):
    theta = 0
    width_x = 0.2
    width_y = 0.2

    center = np.array(center)

    x = np.linspace(-1, 1, 100) * np.sqrt(width_x)
    y = np.linspace(-1, 1, 100) * np.sqrt(width_y)
    xx, yy = np.meshgrid(x, y)

    # Apply local transformation
    rotation_matrix = np.array(
        [[np.cos(-theta), np.sin(-theta)], [-np.sin(-theta), np.cos(-theta)]]
    )
    xx, yy = transform_pattern(xx, yy, rotation_matrix, [0, 0])
    xx, yy = translate_pattern(xx, yy, -center)

    # Apply global transformation
    if transformation_matrix is not None:
        inv_transformation_matrix = np.linalg.inv(transformation_matrix)
        xx, yy = transform_pattern(xx, yy, inv_transformation_matrix, [0, 0])

    w, h = grid_size
    ar = h / w

    return xx, yy * ar


### Visualization
def plot_grid_points(xx, yy, ax=None, cmap="inferno", scatter_kws=None):
    if ax is None:
        ax = plt.subplots()[1]
    if scatter_kws is not None:
        ax.scatter(xx.flatten(), yy.flatten(), **scatter_kws)
    else:
        h, w = xx.shape
        x = np.linspace(-1, 1, w)
        y = np.linspace(-1, 1, h)
        xx_c, yy_c = np.meshgrid(x, y)
        norm = np.sqrt(((xx_c + 1) / 2) ** 2 + ((yy_c + 1) / 2) ** 2)
        norm = norm / norm.max()
        ax.scatter(xx.flatten(), yy.flatten(), c=norm.flatten(), cmap=cmap)


def plot_grid_border(xx, yy, ax=None, plot_kws=None):
    if ax is None:
        ax = plt.subplots()[1]
    h, w = xx.shape
    first = [0, 0, h - 1, h - 1, 0]
    second = [0, w - 1, w - 1, 0, 0]
    if plot_kws is not None:
        ax.plot(xx[first, second], yy[first, second], **plot_kws)
    else:
        ax.plot(xx[first, second], yy[first, second], c="k", lw=4)


### Generating random transformation matrices
def rotation_matrix(theta):
    return np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])


def scaling_matrix(s_x, s_y):
    return np.array([[s_x, 0], [0, s_y]])


def shearing_matrix(s_h, s_v):
    return np.array([[1, s_h], [s_v, 1]])


def random_transformation_matrices(n):
    matrices = []
    for _ in range(n):
        theta = np.random.uniform(0, 2 * np.pi)  # Random rotation
        s_x = np.random.uniform(0.5, 1.3)  # Random scaling factors
        s_y = s_x + np.random.uniform(-0.3, 0.3)
        s_h = np.random.uniform(-0.15, 0.15)  # Random horizontal shearing
        s_v = np.random.uniform(-0.15, 0.15)  # Random vertical shearing

        rotation = rotation_matrix(theta)
        scaling = scaling_matrix(s_x, s_y)
        shearing = shearing_matrix(s_h, s_v)

        transformation = np.dot(rotation, np.dot(scaling, shearing))
        matrices.append(transformation)

    return matrices



import logging
import warnings
from typing import Tuple, Union

import numpy as np
from numpy.typing import ArrayLike

from ..utils import anscombe

logger = logging.getLogger(__name__)


def corr(
    y1: ArrayLike, y2: ArrayLike, axis: Union[None, int, Tuple[int]] = -1, eps: int = 1e-8, **kwargs
) -> np.ndarray:
    """
    Compute the correlation between two NumPy arrays along the specified dimension(s).

    Args:
        y1:      first NumPy array
        y2:      second NumPy array
        axis:    dimension(s) along which the correlation is computed. Any valid NumPy axis spec works here
        eps:     offset to the standard deviation to avoid exploding the correlation due to small division (default 1e-8)
        **kwargs: passed to final numpy.mean operation over standardized y1 * y2

    Returns: correlation array
    """

    y1 = (y1 - y1.mean(axis=axis, keepdims=True)) / (y1.std(axis=axis, keepdims=True, ddof=0) + eps)
    y2 = (y2 - y2.mean(axis=axis, keepdims=True)) / (y2.std(axis=axis, keepdims=True, ddof=0) + eps)
    return (y1 * y2).mean(axis=axis, **kwargs)


def oracle_corr_conservative(repeated_outputs: ArrayLike) -> np.ndarray:
    """
    Compute the corrected oracle correlations per neuron.
    Note that an unequal number of repeats will introduce bias as it distorts assumptions made about the dataset.
    Note that oracle_corr_conservative overestimates the true oracle correlation.

    Args:
        repeated_outputs (array-like): numpy array with shape (images, repeats, neuron responses), or a list containing for each
            image an array of shape (repeats, neuron responses).

    Returns:
        array: Corrected oracle correlations per neuron
    """

    var_noise, var_mean = [], []
    for output in repeated_outputs:
        var_noise.append(output.var(axis=0))
        var_mean.append(output.mean(axis=0))
    var_noise = np.mean(np.array(var_noise), axis=0)
    var_mean = np.var(np.array(var_mean), axis=0)
    return var_mean / np.sqrt(var_mean * (var_mean + var_noise))


def oracle_corr_jackknife(repeated_outputs: ArrayLike) -> np.ndarray:
    """
    Compute the oracle correlations per neuron.
    Note that an unequal number of repeats will introduce bias as it distorts assumptions made about the dataset.
    Note that oracle_corr_jackknife underestimates the true oracle correlation.

    Args:
        repeated_outputs (array-like): numpy array with shape (images, repeats, neuron responses), or a list containing for each
            image an array of shape (repeats, neuron responses).

    Returns:
        array: Oracle correlations per neuron
    """

    oracles = []
    for outputs in repeated_outputs:
        num_repeats = outputs.shape[0]
        oracle = (outputs.sum(axis=0, keepdims=True) - outputs) / (num_repeats - 1)
        if np.any(np.isnan(oracle)):
            logger.warning(
                "{}% NaNs when calculating the oracle. NaNs will be set to Zero.".format(np.isnan(oracle).mean() * 100)
            )
            oracle[np.isnan(oracle)] = 0
        oracles.append(oracle)
    return corr(np.vstack(repeated_outputs), np.vstack(oracles), axis=0)


def explainable_var(repeated_outputs: ArrayLike, eps: int = 1e-9) -> np.ndarray:
    """
    Compute the explainable variance per neuron.

    Args:
        repeated_outputs (array): numpy array with shape (images, repeats, neuron responses), or a list containing for each
            image an array of shape (repeats, neuron responses).

    Returns:
        array: Corrected oracle correlations per neuron
    """

    total_var = np.var(np.vstack(repeated_outputs), axis=0, ddof=1)
    img_var = np.var(repeated_outputs, axis=1, ddof=1)
    noise_var = np.mean(img_var, axis=0)
    explainable_var = (total_var - noise_var) / (total_var + eps)
    return explainable_var


def fev(targets: ArrayLike, predictions: ArrayLike, return_exp_var: bool = False) -> Union[ArrayLike, Tuple[ArrayLike]]:
    """
    Compute the fraction of explainable variance explained per neuron

    Args:
        targets (array-like): Neuronal neuron responses (ground truth) to image repeats. Dimensions:
            [num_images] np.array(num_repeats, num_neurons)
        outputs (array-like): Model predictions to the repeated images, with an identical shape as the targets
        return_exp_var (bool): returns the fraction of explainable variance per neuron if set to True
    Returns:
        FEVe (np.array): the fraction of explainable variance explained per neuron
        --- optional: FEV (np.array): the fraction
    """

    img_var = []
    pred_var = []
    for target, prediction in zip(targets, predictions):
        pred_var.append((target - prediction) ** 2)
        img_var.append(np.var(target, axis=0, ddof=1))
    pred_var = np.vstack(pred_var)
    img_var = np.vstack(img_var)

    total_var = np.var(np.vstack(targets), axis=0, ddof=1)
    noise_var = np.mean(img_var, axis=0)
    fev = (total_var - noise_var) / total_var

    pred_var = np.mean(pred_var, axis=0)
    fev_e = 1 - (pred_var - noise_var) / (total_var - noise_var)
    return [fev, fev_e] if return_exp_var else fev_e


def snr(repeated_outputs: ArrayLike, per_neuron: bool = True) -> np.ndarray:
    """
    Compute signal to noise ratio.

    Args:
        repeated_outputs (array-like): numpy array with shape (images, repeats, neuron responses), or a list containing for each
            image an array of shape (repeats, neuron responses).
        per_neuron (bool, optional): Return snr per neuron or averaged across neurons. Defaults to True.

    Returns:
        array: Signal to noise ratio per neuron or averaged across neurons.
    """
    repeated_outputs = anscombe(repeated_outputs)
    mu = np.array([np.mean(repeats, axis=0) for repeats in repeated_outputs])
    mu_bar = np.mean(mu, axis=0)
    sigma_2 = np.array([np.var(repeats, ddof=1, axis=0) for repeats in repeated_outputs])
    sigma_2_bar = np.mean(sigma_2, axis=0)
    snr = (1 / mu.shape[0] * np.sum((mu - mu_bar) ** 2, axis=0)) / sigma_2_bar
    return snr if per_neuron else np.mean(snr)


def gini(x, axis=None):
    """
    Calculate the Gini coefficient from a list of numbers. The Gini coefficient is used as a measure of (in)equality
    where a Gini coefficient of 1 (or 100%) expresses maximal inequality among values. A value greater than 1 may occur
     if some value represents negative contribution.

    Args:
        x: 1 D array or list
            Array of numbers from which to calculate the Gini coefficient.
        axis: axis along which to compute gini. If None, then the array is flattened first.

    Returns: float
            Gini coefficient
    """
    x = np.asarray(x)
    if axis is None:
        x = x.flatten()
        axis = -1
    if np.any(x < 0):
        warnings.warn("Input x contains negative values")
    sorted_x = np.sort(x, axis=axis)
    n = x.shape[axis]
    cumx = np.cumsum(sorted_x, dtype=float, axis=axis)
    return (n + 1 - 2 * np.sum(cumx, axis=axis) / cumx.take(-1, axis=axis)) / n



import warnings
from contextlib import contextmanager

import torch


@contextmanager
def eval_state(model):
    """
    Context manager, within which the model will be under `eval` mode.
    Upon existing, the model will return to whatever training state it
    was as it entered into the context.

    Args:
        model (PyTorch Module): PyTorch Module whose train/eval state is to be managed.

    Yields:
        PyTorch Module: The model switched to eval state.
    """
    training_status = model.training

    try:
        model.eval()
        yield model
    finally:
        model.train(training_status)


@contextmanager
def device_state(model, device):
    """
    Within the context, attemps to place the `model` onto the specified
    `device`. If `device` is CUDA and the specified device does not exist,
    the context falls back to using `cpu`. Upon existing the context, the model
    will be placed back on to the original device inferred based on the first entry
    of the model's parameter.

    Args:
        model (PyTorch Module): PyTorch Module object to swtich device.
        device (Any): target device descriptor. Any valid PyTorch device descriptor may be used.

    Yields:
        PyTorch Module: Model placed on the new device
    """
    # infer the original device based on the device the first parameter is placed on
    original_device = next(model.parameters()).device

    # create device spec
    # if device is simply "cuda", then device.index will evaluate to one, and the if statement will error out.
    device = torch.device("cuda:0") if device == "cuda" else torch.device(device)
    if device.type == "cuda" and device.index >= torch.cuda.device_count():
        # fall back to using CPU
        warnings.warn("Incompatible CUDA spec. Falling back to CPU usage")
        device = "cpu"

    try:
        model.to(device)
        yield model
    finally:
        model.to(original_device)



from collections import OrderedDict

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import random
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from neuralpredictors.layers.attention import AttentionConv
from neuralpredictors.layers.conv import DepthSeparableConv2d
from neuralpredictors.layers.cores.base import Core
from neuralpredictors import regularizers
from neuralpredictors.layers.readouts import PointPooled2d, FullGaussian2d
from neuralpredictors.utils import get_module_output


def load_stored(model_class, path, map_location=None):
    hparams_and_state_dict = torch.load(path, map_location)
    model = eval(model_class)(**hparams_and_state_dict["hparams"])
    model.load_state_dict(hparams_and_state_dict["model_state_dict"])
    return model


def get_io_dims(data_loader):
    """
    Returns the shape of the dataset for each item within an entry returned by the `data_loader`
    The DataLoader object must return either a namedtuple, dictionary or a plain tuple.
    If `data_loader` entry is a namedtuple or a dictionary, a dictionary with the same keys as the
    namedtuple/dict item is returned, where values are the shape of the entry. Otherwise, a tuple of
    shape information is returned.

    Note that the first dimension is always the batch dim with size depending on the data_loader configuration.

    Args:
        data_loader (torch.DataLoader): is expected to be a pytorch Dataloader object returning
            either a namedtuple, dictionary, or a plain tuple.
    Returns:
        dict or tuple: If data_loader element is either namedtuple or dictionary, a ditionary
            of shape information, keyed for each entry of dataset is returned. Otherwise, a tuple
            of shape information is returned. The first dimension is always the batch dim
            with size depending on the data_loader configuration.
    """
    items = next(iter(data_loader))
    if hasattr(items, "_asdict"):  # if it's a named tuple
        items = items._asdict()

    if hasattr(items, "items"):  # if dict like
        return {k: v.shape for k, v in items.items()}
    else:
        return (v.shape for v in items)


def unpack_data_info(data_info):
    in_shapes_dict = {k: v["input_dimensions"] for k, v in data_info.items()}
    input_channels = [v["input_channels"] for k, v in data_info.items()]
    n_neurons_dict = {k: v["output_dimension"] for k, v in data_info.items()}
    return n_neurons_dict, in_shapes_dict, input_channels


def set_random_seed(seed: int, deterministic: bool = True):
    """
    Set random generator seed for Python interpreter, NumPy and PyTorch. When setting the seed for PyTorch,
    if CUDA device is available, manual seed for CUDA will also be set. Finally, if `deterministic=True`,
    and CUDA device is available, PyTorch CUDNN backend will be configured to `benchmark=False` and `deterministic=True`
    to yield as deterministic result as possible. For more details, refer to
    PyTorch documentation on reproducibility: https://pytorch.org/docs/stable/notes/randomness.html

    Beware that the seed setting is a "best effort" towards deterministic run. However, as detailed in the above documentation,
    there are certain PyTorch CUDA opertaions that are inherently non-deterministic, and there is no simple way to control for them.
    Thus, it is best to assume that when CUDA is utilized, operation of the PyTorch module will not be deterministic and thus
    not completely reproducible.

    Args:
        seed (int): seed value to be set
        deterministic (bool, optional): If True, CUDNN backend (if available) is set to be deterministic. Defaults to True. Note that if set
            to False, the CUDNN properties remain untouched and it NOT explicitly set to False.
    """
    random.seed(seed)
    np.random.seed(seed)
    if deterministic:
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
    torch.manual_seed(seed)  # this sets both CPU and CUDA seeds for PyTorch


class SQ_EX_Block(nn.Module):
    def __init__(self, in_ch, reduction=16):
        super(SQ_EX_Block, self).__init__()
        self.se = nn.Sequential(
            GlobalAvgPool(),
            nn.Linear(in_ch, in_ch // reduction),
            nn.ReLU(inplace=True),
            nn.Linear(in_ch // reduction, in_ch),
            nn.Sigmoid(),
        )

    def forward(self, x):
        se_weight = self.se(x).unsqueeze(-1).unsqueeze(-1)
        return x.mul(se_weight)


class GlobalAvgPool(nn.Module):
    def __init__(self):
        super(GlobalAvgPool, self).__init__()

    def forward(self, x):
        return x.view(*(x.shape[:-2]), -1).mean(-1)


class SE2dCore(Core, nn.Module):
    def __init__(
        self,
        input_channels,
        hidden_channels,
        input_kern,
        hidden_kern,
        layers=3,
        gamma_input=0.0,
        gamma_hidden=0.0,
        skip=0,
        final_nonlinearity=True,
        final_batch_norm=True,
        bias=False,
        momentum=0.1,
        pad_input=True,
        batch_norm=True,
        hidden_dilation=1,
        laplace_padding=None,
        input_regularizer="LaplaceL2norm",
        stack=None,
        se_reduction=32,
        n_se_blocks=1,
        depth_separable=False,
        attention_conv=False,
        linear=False,
        first_layer_stride=1,
    ):
        """
        Args:
            input_channels:     Integer, number of input channels as in
            hidden_channels:    Number of hidden channels (i.e feature maps) in each hidden layer
            input_kern:     kernel size of the first layer (i.e. the input layer)
            hidden_kern:    kernel size of each hidden layer's kernel
            layers:         number of layers
            gamma_input:    regularizer factor for the input weights (default: LaplaceL2, see neuralpredictors.regularizers)
            skip:           Adds a skip connection
            final_nonlinearity: Boolean, if true, appends an ELU layer after the last BatchNorm (if BN=True)
            bias:           Adds a bias layer. Note: bias and batch_norm can not both be true
            momentum:       BN momentum
            pad_input:      Boolean, if True, applies zero padding to all convolutions
            batch_norm:     Boolean, if True appends a BN layer after each convolutional layer
            hidden_dilation:    If set to > 1, will apply dilated convs for all hidden layers
            laplace_padding: Padding size for the laplace convolution. If padding = None, it defaults to half of
                the kernel size (recommended). Setting Padding to 0 is not recommended and leads to artefacts,
                zero is the default however to recreate backwards compatibility.
            normalize_laplace_regularizer: Boolean, if set to True, will use the LaplaceL2norm function from
                neuralpredictors.regularizers, which returns the regularizer as |laplace(filters)| / |filters|
            input_regularizer:  String that must match one of the regularizers in ..regularizers
            stack:        Int or iterable. Selects which layers of the core should be stacked for the readout.
                            default value will stack all layers on top of each other.
                            stack = -1 will only select the last layer as the readout layer
                            stack = 0  will only readout from the first layer
            se_reduction:   Int. Reduction of Channels for Global Pooling of the Squeeze and Excitation Block.
            attention_conv: Boolean, if True, uses self-attention instead of convolution for layers 2 and following
        """

        super().__init__()

        assert not bias or not batch_norm, "bias and batch_norm should not both be true"
        assert not depth_separable or not attention_conv, (
            "depth_separable and attention_conv should not both be true"
        )

        regularizer_config = (
            dict(padding=laplace_padding, kernel=input_kern)
            if input_regularizer == "GaussianLaplaceL2"
            else dict(padding=laplace_padding)
        )
        self._input_weights_regularizer = regularizers.__dict__[input_regularizer](
            **regularizer_config
        )

        self.layers = layers
        self.depth_separable = depth_separable
        self.gamma_input = gamma_input
        self.gamma_hidden = gamma_hidden
        self.input_channels = input_channels
        self.hidden_channels = hidden_channels
        self.skip = skip
        self.features = nn.Sequential()
        self.n_se_blocks = n_se_blocks
        if stack is None:
            self.stack = range(self.layers)
        else:
            self.stack = (
                [*range(self.layers)[stack:]] if isinstance(stack, int) else stack
            )

        # --- first layer
        layer = OrderedDict()
        layer["conv"] = nn.Conv2d(
            input_channels,
            hidden_channels,
            input_kern,
            padding=input_kern // 2 if pad_input else 0,
            bias=bias,
            stride=first_layer_stride,
        )
        if batch_norm:
            layer["norm"] = nn.BatchNorm2d(hidden_channels, momentum=momentum)
        if (layers > 1 or final_nonlinearity) and not linear:
            layer["nonlin"] = nn.ELU(inplace=True)
        self.features.add_module("layer0", nn.Sequential(layer))

        if not isinstance(hidden_kern, Iterable):
            hidden_kern = [hidden_kern] * (self.layers - 1)

        # --- other layers
        for l in range(1, self.layers):
            layer = OrderedDict()
            hidden_padding = ((hidden_kern[l - 1] - 1) * hidden_dilation + 1) // 2
            if depth_separable:
                layer["ds_conv"] = DepthSeparableConv2d(
                    hidden_channels,
                    hidden_channels,
                    kernel_size=hidden_kern[l - 1],
                    dilation=hidden_dilation,
                    padding=hidden_padding,
                    bias=False,
                    stride=1,
                )
            elif attention_conv:
                layer["conv"] = AttentionConv(
                    hidden_channels if not skip > 1 else min(skip, l) * hidden_channels,
                    hidden_channels,
                    hidden_kern[l - 1],
                    padding=hidden_padding,
                    bias=bias and not batch_norm,
                )
            else:
                layer["conv"] = nn.Conv2d(
                    hidden_channels if not skip > 1 else min(skip, l) * hidden_channels,
                    hidden_channels,
                    hidden_kern[l - 1],
                    padding=hidden_padding,
                    bias=bias,
                    dilation=hidden_dilation,
                )
            if (final_batch_norm or l < self.layers - 1) and batch_norm:
                layer["norm"] = nn.BatchNorm2d(hidden_channels, momentum=momentum)

            if (final_nonlinearity or l < self.layers - 1) and not linear:
                layer["nonlin"] = nn.ELU(inplace=True)

            if (self.layers - l) <= self.n_se_blocks:
                layer["seg_ex_block"] = SQ_EX_Block(
                    in_ch=hidden_channels, reduction=se_reduction
                )

            self.features.add_module("layer{}".format(l), nn.Sequential(layer))

        self.apply(self.init_conv)

    def forward(self, input_):
        ret = []
        for l, feat in enumerate(self.features):
            do_skip = l >= 1 and self.skip > 1
            input_ = feat(
                input_ if not do_skip else torch.cat(ret[-min(self.skip, l) :], dim=1)
            )
            if l in self.stack:
                ret.append(input_)
        return torch.cat(ret, dim=1)

    def laplace(self):
        return self._input_weights_regularizer(self.features[0].conv.weight)

    def group_sparsity(self):
        ret = 0
        for l in range(1, self.layers):
            if self.depth_separable:
                for ds_i in range(3):
                    ret = (
                        ret
                        + self.features[l]
                        .ds_conv[ds_i]
                        .weight.pow(2)
                        .sum(3, keepdim=True)
                        .sum(2, keepdim=True)
                        .sqrt()
                        .mean()
                    )
            else:
                ret = (
                    ret
                    + self.features[l]
                    .conv.weight.pow(2)
                    .sum(3, keepdim=True)
                    .sum(2, keepdim=True)
                    .sqrt()
                    .mean()
                )
        return ret / ((self.layers - 1) if self.layers > 1 else 1)

    def regularizer(self):
        return (
            self.gamma_input * self.laplace()
            + self.gamma_hidden * self.group_sparsity()
        )

    @property
    def outchannels(self):
        return len(self.features) * self.hidden_channels


class MultiReadout:
    def forward(self, *args, data_key=None, **kwargs):
        if data_key is None and len(self) == 1:
            data_key = list(self.keys())[0]
        return self[data_key](*args, **kwargs)

    def regularizer(self, data_key):
        return self[data_key].feature_l1(average=False) * self.gamma_readout


class MultiplePointPooled2d(MultiReadout, torch.nn.ModuleDict):
    def __init__(
        self,
        core,
        in_shape_dict,
        n_neurons_dict,
        pool_steps,
        pool_kern,
        bias,
        init_range,
        gamma_readout,
    ):
        # super init to get the _module attribute
        super().__init__()
        for k in n_neurons_dict:
            in_shape = get_module_output(core, in_shape_dict[k])[1:]
            n_neurons = n_neurons_dict[k]
            self.add_module(
                k,
                PointPooled2d(
                    in_shape,
                    n_neurons,
                    pool_steps=pool_steps,
                    pool_kern=pool_kern,
                    bias=bias,
                    init_range=init_range,
                ),
            )
        self.gamma_readout = gamma_readout


class MultipleFullGaussian2d(nn.ModuleDict):
    def __init__(
        self,
        core,
        in_shape_dict,
        n_neurons_dict,
        init_mu_range,
        init_sigma_range,
        bias,
        gamma_readout,
    ):
        # super init to get the _module attribute
        super(MultipleFullGaussian2d, self).__init__()
        for k in n_neurons_dict:
            in_shape = get_module_output(core, in_shape_dict[k])[1:]
            n_neurons = n_neurons_dict[k]
            self.add_module(
                k,
                FullGaussian2d(
                    in_shape=in_shape,
                    outdims=n_neurons,
                    init_mu_range=init_mu_range,
                    init_sigma_range=init_sigma_range,
                    bias=bias,
                ),
            )

        self.gamma_readout = gamma_readout

    def forward(self, *args, data_key=None, **kwargs):
        if data_key is None and len(self) == 1:
            data_key = list(self.keys())[0]
        return self[data_key](*args, **kwargs)

    def regularizer(self, data_key):
        return self[data_key].feature_l1(average=False) * self.gamma_readout


def se_core_point_readout(
    data_info,
    seed,
    hidden_channels=32,
    input_kern=13,  # core args
    hidden_kern=3,
    layers=3,
    gamma_input=15.5,
    skip=0,
    final_nonlinearity=True,
    momentum=0.9,
    pad_input=False,
    batch_norm=True,
    hidden_dilation=1,
    laplace_padding=None,
    input_regularizer="LaplaceL2norm",
    pool_steps=2,
    pool_kern=3,
    init_range=0.2,
    readout_bias=True,  # readout args,
    gamma_readout=4,
    elu_offset=0,
    stack=None,
    se_reduction=32,
    n_se_blocks=1,
    depth_separable=False,
    linear=False,
):
    """
    Model class of a stacked2dCore (from neuralpredictors) and a pointpooled (spatial transformer) readout

    Args:
        dataloaders: a dictionary of dataloaders, one loader per session
            in the format {'data_key': dataloader object, .. }
        seed: random seed
        elu_offset: Offset for the output non-linearity [F.elu(x + self.offset)]

        all other args: See Documentation of Stacked2dCore in neuralpredictors.layers.cores and
            PointPooled2D in neuralpredictors.layers.readouts

    Returns: An initialized model which consists of model.core and model.readout
    """
    n_neurons_dict, in_shapes_dict, input_channels = unpack_data_info(data_info)
    core_input_channels = (
        list(input_channels.values())[0]
        if isinstance(input_channels, dict)
        else input_channels[0]
    )

    class Encoder(nn.Module):
        def __init__(self, core, readout, elu_offset):
            super().__init__()
            self.core = core
            self.readout = readout
            self.offset = elu_offset

        def forward(self, x, data_key=None, **kwargs):
            x = self.core(x)

            x = self.readout(x, data_key=data_key)
            return F.elu(x + self.offset) + 1

        def regularizer(self, data_key):
            return self.core.regularizer() + self.readout.regularizer(data_key=data_key)

    set_random_seed(seed)

    core = SE2dCore(
        input_channels=core_input_channels,
        hidden_channels=hidden_channels,
        input_kern=input_kern,
        hidden_kern=hidden_kern,
        layers=layers,
        gamma_input=gamma_input,
        skip=skip,
        final_nonlinearity=final_nonlinearity,
        bias=False,
        momentum=momentum,
        pad_input=pad_input,
        batch_norm=batch_norm,
        hidden_dilation=hidden_dilation,
        laplace_padding=laplace_padding,
        input_regularizer=input_regularizer,
        stack=stack,
        se_reduction=se_reduction,
        n_se_blocks=n_se_blocks,
        depth_separable=depth_separable,
        linear=linear,
    )

    readout = MultiplePointPooled2d(
        core,
        in_shape_dict=in_shapes_dict,
        n_neurons_dict=n_neurons_dict,
        pool_steps=pool_steps,
        pool_kern=pool_kern,
        bias=readout_bias,
        gamma_readout=gamma_readout,
        init_range=init_range,
    )
    model = Encoder(core, readout, elu_offset)
    return model


def se_core_full_gauss_readout(
    data_info,
    seed,
    hidden_channels=32,
    input_kern=13,
    hidden_kern=3,
    layers=3,
    gamma_input=15.5,
    skip=0,
    final_nonlinearity=True,
    momentum=0.9,
    pad_input=False,
    batch_norm=True,
    hidden_dilation=1,
    laplace_padding=None,
    input_regularizer="LaplaceL2norm",
    init_mu_range=0.2,
    init_sigma=1.0,
    readout_bias=True,
    gamma_readout=4,
    elu_offset=0,
    stack=None,
    se_reduction=32,
    n_se_blocks=1,
    depth_separable=False,
    linear=False,
):
    """
    Model class of a stacked2dCore (from neuralpredictors) and a pointpooled (spatial transformer) readout

    Args:
        dataloaders: a dictionary of dataloaders, one loader per session
            in the format {'data_key': dataloader object, .. }
        seed: random seed
        elu_offset: Offset for the output non-linearity [F.elu(x + self.offset)]
        isotropic: whether the Gaussian readout should use isotropic Gaussians or not
        grid_mean_predictor: if not None, needs to be a dictionary of the form
            {
            'type': 'cortex',
            'input_dimensions': 2,
            'hidden_layers':0,
            'hidden_features':20,
            'final_tanh': False,
            }
            In that case the datasets need to have the property `neurons.cell_motor_coordinates`
        share_features: whether to share features between readouts. This requires that the datasets
            have the properties `neurons.multi_match_id` which are used for matching. Every dataset
            has to have all these ids and cannot have any more.
        all other args: See Documentation of Stacked2dCore in neuralpredictors.layers.cores and
            PointPooled2D in neuralpredictors.layers.readouts

    Returns: An initialized model which consists of model.core and model.readout
    """
    n_neurons_dict, in_shapes_dict, input_channels = unpack_data_info(data_info)
    core_input_channels = (
        list(input_channels.values())[0]
        if isinstance(input_channels, dict)
        else input_channels[0]
    )

    class Encoder(nn.Module):
        def __init__(self, core, readout, elu_offset):
            super().__init__()
            self.core = core
            self.readout = readout
            self.offset = elu_offset

        def forward(self, x, data_key=None, **kwargs):
            x = self.core(x)

            sample = kwargs["sample"] if "sample" in kwargs else None
            x = self.readout(x, data_key=data_key, sample=sample)
            return F.elu(x + self.offset) + 1

        def regularizer(self, data_key):
            return self.core.regularizer() + self.readout.regularizer(data_key=data_key)

    set_random_seed(seed)

    core = SE2dCore(
        input_channels=core_input_channels,
        hidden_channels=hidden_channels,
        input_kern=input_kern,
        hidden_kern=hidden_kern,
        layers=layers,
        gamma_input=gamma_input,
        skip=skip,
        final_nonlinearity=final_nonlinearity,
        bias=False,
        momentum=momentum,
        pad_input=pad_input,
        batch_norm=batch_norm,
        hidden_dilation=hidden_dilation,
        laplace_padding=laplace_padding,
        input_regularizer=input_regularizer,
        stack=stack,
        se_reduction=se_reduction,
        n_se_blocks=n_se_blocks,
        depth_separable=depth_separable,
        linear=linear,
    )

    readout = MultipleFullGaussian2d(
        core,
        in_shape_dict=in_shapes_dict,
        n_neurons_dict=n_neurons_dict,
        init_mu_range=init_mu_range,
        bias=readout_bias,
        init_sigma_range=init_sigma,
        gamma_readout=gamma_readout,
    )

    model = Encoder(core, readout, elu_offset)

    return model



"""
For most of our neural prediction models, we use a core and readout architecture.
In this module, all the cores can be found.
The core's task is to encode the input into (hidden) representations, which are shared across all neurons.
These representations will be passed on to the readout to calculate the desired output.

Our cores are usually CNNs. The core module is using special architectures submodules from ..layers (such as
attention_conv which uses self attention instead a conv2d layer).

All core classes must have a `regularizer` and a `forward` method.
"""

from .conv2d import (
    RotationEquivariant2dCore,
    SE2dCore,
    Stacked2dCore,
    TransferLearningCore,
)



from collections import namedtuple

import h5py
import numpy as np
from scipy.signal import convolve2d

from ..transforms import DataTransform, Delay, MovieTransform, Subsequence
from ..utils import recursively_load_dict_contents_from_group
from .base import (
    AttributeTransformer,
    DirectoryAttributeHandler,
    FileTreeDatasetBase,
    TransformDataset,
)


class H5SequenceSet(TransformDataset):
    def __init__(self, filename, *data_keys, output_rename=None, transforms=None, output_dict=False):
        super().__init__(transforms=transforms)

        self.output_dict = output_dict

        if output_rename is None:
            output_rename = {}

        # a flag that can be changed to turn renaming on/off
        self.rename_output = True

        self.output_rename = output_rename

        self._fid = h5py.File(filename, "r")
        self.data = self._fid
        self.data_loaded = False

        # ensure that all elements of data_keys exist
        m = None
        for key in data_keys:
            assert key in self.data, "Could not find {} in file".format(key)
            l = len(self.data[key])
            if m is not None and l != m:
                raise ValueError("groups have different length")
            m = l
        self._len = m

        # Specify which types of transforms are accepted
        self._transform_set = DataTransform

        self.data_keys = data_keys
        self.transforms = transforms or []

        self.data_point = namedtuple("DataPoint", data_keys)
        renamed_keys = [output_rename.get(k, k) for k in data_keys]
        self.output_point = namedtuple("OutputPoint", renamed_keys)

    def __dir__(self):
        attrs = set(super().__dir__())
        return attrs.union(set(self._fid.keys()))

    def load_content(self):
        self.data = recursively_load_dict_contents_from_group(self._fid)
        self.data_loaded = True

    def unload_content(self):
        self.data = self._fid
        self.data_loaded = False

    def __len__(self):
        return self._len

    def __getitem__(self, item):
        x = self.data_point(*(np.array(self.data[g][item if self.data_loaded else str(item)]) for g in self.data_keys))
        for tr in self.transforms:
            assert isinstance(tr, self._transform_set)
            x = tr(x)

        # convert to output point
        if self.rename_output:
            x = self.output_point(*x)

        if self.output_dict:
            x = x._asdict()
        return x

    def __getattr__(self, item):
        if item in self.data:
            item = self.data[item]
            if isinstance(item, h5py.Dataset):
                dtype = item.dtype
                item = item[()]
                if dtype.char == "S":  # convert bytes to unicode
                    item = item.astype(str)
                return item
            return item
        else:
            # TODO: check for a proper way to handle cases where super doesn't have __getattr__
            return super().__getattr__(item)

    def __repr__(self):
        names = ["{} -> {}".format(k, self.output_rename[k]) if k in self.output_rename else k for k in self.data_keys]
        s = "{} m={}:\n\t({})".format(self.__class__.__name__, len(self), ", ".join(names))
        if self.transforms is not None:
            s += "\n\t[Transforms: " + "->".join([repr(tr) for tr in self.transforms]) + "]"
        return s


class MovieSet(H5SequenceSet):
    """
    Extension to H5SequenceSet with specific HDF5 dataset assumed. Specifically,
    it assumes that properties such as `neurons` and `stats` are present in the dataset.
    """

    def __init__(self, filename, *data_groups, output_rename=None, transforms=None, stats_source="all"):
        super().__init__(filename, *data_groups, output_rename=output_rename, transforms=transforms)
        self.stats_source = stats_source

        # set to accept only MovieTransform
        self._transform_set = MovieTransform

    @property
    def neurons(self):
        return AttributeTransformer("neurons", self.data, self.transforms, data_group="responses")

    @property
    def n_neurons(self):
        return len(self.neurons.unit_ids)

    @property
    def input_shape(self):
        name = self.output_rename.get("inputs", "inputs") if self.rename_output else "inputs"
        return (1,) + getattr(self[0], name).shape

    def transformed_mean(self, stats_source=None):
        if stats_source is None:
            stats_source = self.stats_source

        tmp = [np.atleast_1d(self.statistics[g][stats_source]["mean"][()]) for g in self.data_keys]
        x = self.transform(self.data_point(*tmp), exclude=(Subsequence, Delay))
        if self.rename_output:
            x = self.output_point(*x)
        return x

    def rf_base(self, stats_source="all"):
        N, c, t, w, h = self.img_shape
        t = min(t, 150)
        mean = lambda dk: self.statistics[dk][stats_source]["mean"][()]
        d = dict(
            inputs=np.ones((1, c, t, w, h)) * np.array(mean("inputs")),
            eye_position=np.ones((1, t, 1)) * mean("eye_position")[None, None, :],
            behavior=np.ones((1, t, 1)) * mean("behavior")[None, None, :],
            responses=np.ones((1, t, 1)) * mean("responses")[None, None, :],
        )
        return self.transform(self.data_point(*[d[dk] for dk in self.data_keys]), exclude=Subsequence)

    def rf_noise_stim(self, m, t, stats_source="all"):
        """
        Generates a Gaussian white noise stimulus filtered with a 3x3 Gaussian filter
        for the computation of receptive fields. The mean and variance of the Gaussian
        noise are set to the mean and variance of the stimulus ensemble.
        The behvavior, eye movement statistics, and responses are set to their respective means.
        Args:
            m: number of noise samples
            t: length in time
        Returns: tuple of input, behavior, eye, and response
        """
        N, c, _, w, h = self.img_shape
        stat = lambda dk, what: self.statistics[dk][stats_source][what][()]
        mu, s = stat("inputs", "mean"), stat("inputs", "std")
        h_filt = np.float64([[1 / 16, 1 / 8, 1 / 16], [1 / 8, 1 / 4, 1 / 8], [1 / 16, 1 / 8, 1 / 16]])
        noise_input = (
            np.stack([convolve2d(np.random.randn(w, h), h_filt, mode="same") for _ in range(m * t * c)]).reshape(
                (m, c, t, w, h)
            )
            * s
            + mu
        )

        mean_beh = np.ones((m, t, 1)) * stat("behavior", "mean")[None, None, :]
        mean_eye = np.ones((m, t, 1)) * stat("eye_position", "mean")[None, None, :]
        mean_resp = np.ones((m, t, 1)) * stat("responses", "mean")[None, None, :]

        d = dict(
            inputs=noise_input.astype(np.float32),
            eye_position=mean_eye.astype(np.float32),
            behavior=mean_beh.astype(np.float32),
            responses=mean_resp.astype(np.float32),
        )

        return self.transform(
            self.data_point(*[d[dk] for dk in self.data_groups.values()]), exclude=(Subsequence, Delay)
        )


class MovieFileTreeDataset(FileTreeDatasetBase):
    _transform_types = (MovieTransform,)

    def __init__(self, *args, stats_source=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.stats_source = stats_source if stats_source is not None else "all"

    # the followings are provided for compatibility with MovieSet
    @property
    def types(self):
        return self.trial_info.types

    @property
    def tiers(self):
        return self.trial_info.tiers

    @property
    def n_neurons(self):
        target_group = "responses" if "responses" in self.data_keys else "targets"
        # check if output has been renamed
        if self.rename_output:
            target_group = self._output_rename.get(target_group, target_group)

        val = self[0]
        if hasattr(val, target_group):
            val = getattr(val, target_group)
        else:
            val = val[target_group]
        return val.shape[-1]

    @property
    def statistics(self):
        return DirectoryAttributeHandler(self.basepath / "meta/statistics", self.config["links"])

    def transformed_mean(self, stats_source=None):
        if stats_source is None:
            stats_source = self.stats_source

        tmp = [np.atleast_1d(self.statistics[g][stats_source]["mean"][()]) for g in self.data_keys]
        x = self.transform(self.data_point(*tmp), exclude=(Subsequence, Delay))
        if self.rename_output:
            x = self._output_point(*x)
        return x



import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import numpy as np


def check_activity_requirements(acts, requirements):
    passed = False
    if (
        acts.mean() > requirements["avg"]
        and acts.std() < requirements["std"]
        and acts.min() > requirements["necessary_min"]
    ):
        passed = True
    return passed


class JitteringGridDatamodule:
    def __init__(
        self,
        num_invariances,
        grid_points_per_dim,
        steps_per_epoch,
    ):
        self.num_invariances = num_invariances
        self.grid_points_per_dim = grid_points_per_dim
        self.steps_per_epoch = steps_per_epoch
        self.sigma = 2 * np.pi / self.grid_points_per_dim
        grid = (
            torch.linspace(0, 2 * np.pi, self.grid_points_per_dim + 1)[:-1]
            + self.sigma / 2
        )
        self.grid = torch.stack(
            torch.meshgrid(*[grid for _ in range(self.num_invariances)], indexing='ij'), -1
        ).flatten(0, -2)
        grid0 = torch.linspace(0, 2 * np.pi, self.grid_points_per_dim + 1)[:-1]
        self.grid0 = torch.stack(
            torch.meshgrid(*[grid0 for _ in range(self.num_invariances)], indexing='ij'), -1
        ).flatten(0, -2)
        self.base_grid = self.grid.repeat(self.steps_per_epoch, 1)

    def train_dataloader(self):
        jitter = (
            torch.rand([self.steps_per_epoch, self.num_invariances]) * self.sigma
            - 0.5 * self.sigma
        )
        grids = self.base_grid + jitter.repeat_interleave(
            self.grid_points_per_dim**self.num_invariances, dim=0
        )
        return DataLoader(
            grids,
            batch_size=self.grid_points_per_dim**self.num_invariances,
        )


class ParamReduceOnPlateau:
    def __init__(self, factor=0.8, patience=10, mode="min", threshold=0):
        self.factor = factor
        self.patience = patience
        self.mode = mode
        self.threshold = threshold
        if mode == "min":
            self.op = torch.lt
            self.threshold = -threshold
            self.best_metric = torch.tensor(float('inf'))
        if mode == "max":
            self.op = torch.gt
            self.best_metric = torch.tensor(float('-inf'))
        self.num_epochs_no_improvement = 0

    def step(self, metric_to_monitor, value_to_reduce):
        if self.op(metric_to_monitor, self.best_metric + self.threshold):
            self.best_metric = metric_to_monitor
            self.num_epochs_no_improvement = 0
        else:
            self.num_epochs_no_improvement += 1

        if self.num_epochs_no_improvement > self.patience:
            value_to_reduce = value_to_reduce * self.factor
            self.num_epochs_no_improvement = 0
        return value_to_reduce


class ImprovementChecker:
    def __init__(self, patience=5, ignore_diff_smaller_than=1e-3):
        self.current_value = np.nan
        self.iter = 0
        self.has_not_improved_counter = 0
        self.patience = patience
        self.ignore_diff_smaller_than = ignore_diff_smaller_than

    def is_increasing(self, value):
        if self.iter == 0:
            self.current_value = value
            self.iter += 1
        else:
            if value <= self.current_value + self.ignore_diff_smaller_than:
                self.has_not_improved_counter += 1
            else:
                self.has_not_improved_counter = 0
                self.current_value = value
        return False if self.has_not_improved_counter == self.patience else True



import torch
import torch.nn as nn


def StandardizeClip(mean, std, pixel_min=None, pixel_max=None, dim=[1, 2, 3]):
    transf = [Normalize(mean=mean, std=std, dim=dim)]
    if (pixel_min is not None) or (pixel_max is not None):
        transf.append(Clip(pixel_min=pixel_min, pixel_max=pixel_max))
    return nn.Sequential(*transf)


def FixEnergyNormClip(norm, mean=0, pixel_min=None, pixel_max=None):
    transf = [FixEnergyNorm(norm=norm, mean=mean)]
    if (pixel_min is not None) or (pixel_max is not None):
        transf.append(Clip(pixel_min=pixel_min, pixel_max=pixel_max))
    return nn.Sequential(*transf)


class FixEnergyNorm(nn.Module):
    def __init__(self, norm=0, mean=0, eps=1e-12):
        super().__init__()
        self.zero_mean = mean
        self.norm = norm
        self.eps = eps

    def forward(self, x):
        x_centered = (x - self.zero_mean).view(len(x), -1)
        x_norm2 = torch.sum(x_centered**2, dim=-1, keepdim=True)
        x_norm = torch.sqrt(x_norm2)
        x_renorm = x_centered / (x_norm + self.eps) * self.norm + self.zero_mean
        x_renorm = x_renorm.reshape(x.shape)
        return x_renorm

    def __repr__(self):
        return f"FixEnergyNorm(mean={self.mean}, std={self.std}, dim={self.dim})"


class Normalize(nn.Module):
    def __init__(self, mean=None, std=None, dim=None, eps=1e-12):
        super().__init__()
        self.mean = mean
        self.std = std
        self.dim = dim
        self.eps = eps

    def forward(self, x, iteration=None):
        x_mean = x.mean(dim=self.dim, keepdims=True)
        target_mean = self.mean if self.mean is not None else x_mean
        x_std = x.std(dim=self.dim, keepdims=True)
        target_std = self.std if self.std is not None else x_std
        return target_std * (x - x_mean) / (x_std + self.eps) + target_mean

    def __repr__(self):
        return f"Normalize(mean={self.mean}, std={self.std}, dim={self.dim})"


class Clip(nn.Module):
    def __init__(self, pixel_min, pixel_max):
        super().__init__()
        self.pixel_min = pixel_min
        self.pixel_max = pixel_max

    def forward(self, x):
        x = torch.clamp(x, min=self.pixel_min, max=self.pixel_max)
        return x

    def __repr__(self):
        return f"Clip(min={self.pixel_min}, max={self.pixel_max})"



import warnings
from typing import Any, Literal, Mapping, Optional

import torch
from torch import nn as nn
from torch.nn.modules import Module
from torch.nn.parameter import Parameter

Reduction = Literal["sum", "mean", None]


class ConfigurationError(Exception):
    pass


# ------------------ Base Classes -------------------------


class Readout(Module):
    """
    Base readout class for all individual readouts.
    The MultiReadout will expect its readouts to inherit from this base class.
    """

    features: Parameter
    bias: Parameter

    def initialize(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError("initialize is not implemented for ", self.__class__.__name__)

    def regularizer(self, reduction: Reduction = "sum", average: Optional[bool] = None) -> torch.Tensor:
        raise NotImplementedError("regularizer is not implemented for ", self.__class__.__name__)

    def apply_reduction(
        self, x: torch.Tensor, reduction: Reduction = "mean", average: Optional[bool] = None
    ) -> torch.Tensor:
        """
        Applies a reduction on the output of the regularizer.
        Args:
            x: output of the regularizer
            reduction: method of reduction for the regularizer. Currently possible are ['mean', 'sum', None].
            average: Deprecated. Whether to average the output of the regularizer.
                            If not None, it is transformed into the corresponding value of 'reduction' (see method 'resolve_reduction_method').

        Returns: reduced value of the regularizer
        """
        reduction = self.resolve_reduction_method(reduction=reduction, average=average)

        if reduction == "mean":
            return x.mean()
        elif reduction == "sum":
            return x.sum()
        elif reduction is None:
            return x
        else:
            raise ValueError(
                f"Reduction method '{reduction}' is not recognized. Valid values are ['mean', 'sum', None]"
            )

    def resolve_reduction_method(self, reduction: Reduction = "mean", average: Optional[bool] = None) -> Reduction:
        """
        Helper method which transforms the old and deprecated argument 'average' in the regularizer into
        the new argument 'reduction' (if average is not None). This is done in order to agree with the terminology in pytorch).
        """
        if average is not None:
            warnings.warn("Use of 'average' is deprecated. Please consider using `reduction` instead")
            reduction = "mean" if average else "sum"
        return reduction

    def resolve_deprecated_gamma_readout(self, feature_reg_weight: float, gamma_readout: Optional[float]) -> float:
        if gamma_readout is not None:
            warnings.warn(
                "Use of 'gamma_readout' is deprecated. Please consider using the readout's feature-regularization parameter instead"
            )
            feature_reg_weight = gamma_readout
        return feature_reg_weight

    def initialize_bias(self, mean_activity: Optional[torch.Tensor] = None) -> None:
        """
        Initialize the biases in readout.
        Args:
            mean_activity: Tensor containing the mean activity of neurons.

        Returns:

        """
        if mean_activity is None:
            warnings.warn("Readout is NOT initialized with mean activity but with 0!")
            self.bias.data.fill_(0)
        else:
            self.bias.data = mean_activity

    def __repr__(self) -> str:
        return super().__repr__() + " [{}]\n".format(self.__class__.__name__)  # type: ignore[no-untyped-call,no-any-return]


class ClonedReadout(Module):
    """
    This readout clones another readout while applying a linear transformation on the output. Used for MultiDatasets
    with matched neurons where the x-y positions in the grid stay the same but the predicted responses are rescaled due
    to varying experimental conditions.
    """

    def __init__(self, original_readout: Readout, **kwargs: Any) -> None:
        super().__init__()  # type: ignore[no-untyped-call]

        self._source = original_readout
        self.alpha = Parameter(torch.ones(self._source.features.shape[-1]))  # type: ignore[attr-defined]
        self.beta = Parameter(torch.zeros(self._source.features.shape[-1]))  # type: ignore[attr-defined]

    def forward(self, x: torch.Tensor, **kwarg: Any) -> torch.Tensor:
        x = self._source(x) * self.alpha + self.beta
        return x

    def feature_l1(self, average: bool = True) -> torch.Tensor:
        """Regularization is only applied on the scaled feature weights, not on the bias."""
        if average:
            return (self._source.features * self.alpha).abs().mean()
        else:
            return (self._source.features * self.alpha).abs().sum()

    def initialize(self, **kwargs: Any) -> None:
        self.alpha.data.fill_(1.0)
        self.beta.data.fill_(0.0)



import torch

from .attention import AttentionReadout
from .base import ClonedReadout, Readout
from .factorized import FullFactorized2d, FullSXF, SpatialXFeatureLinear
from .gaussian import (
    DeterministicGaussian2d,
    FullGaussian2d,
    Gaussian2d,
    Gaussian3d,
    RemappedGaussian2d,
    UltraSparse,
)
from .point_pooled import PointPooled2d, SpatialTransformerPooled3d
from .pyramid import PointPyramid2d


class MultiReadoutBase(torch.nn.ModuleDict):
    """
    Base class for MultiReadouts. It is a dictionary of data keys and readouts to the corresponding datasets.
    If parameter-sharing between the readouts is desired, refer to MultiReadoutSharedParametersBase.

    Args:
        in_shape_dict (dict): dictionary of data_key and the corresponding dataset's shape as an output of the core.
        n_neurons_dict (dict): dictionary of data_key and the corresponding dataset's number of neurons
        base_readout (torch.nn.Module): base readout class. If None, self._base_readout must be set manually in the inheriting class's definition
        mean_activity_dict (dict): dictionary of data_key and the corresponding dataset's mean responses. Used to initialize the readout bias with.
                                   If None, the bias is initialized with 0.
        clone_readout (bool): whether to clone the first data_key's readout to all other readouts, only allowing for a scale and offset.
                              This is a rather simple method to enforce parameter-sharing between readouts. For more sophisticated methods,
                              refer to MultiReadoutSharedParametersBase
        gamma_readout (float): regularization strength
        **kwargs:
    """

    _base_readout = None

    def __init__(
        self, in_shape_dict, n_neurons_dict, base_readout=None, mean_activity_dict=None, clone_readout=False, **kwargs
    ):

        # The `base_readout` can be overridden only if the static property `_base_readout` is not set
        if self._base_readout is None:
            self._base_readout = base_readout

        if self._base_readout is None:
            raise ValueError("Attribute _base_readout must be set")
        super().__init__()

        for i, data_key in enumerate(n_neurons_dict):
            first_data_key = data_key if i == 0 else first_data_key
            mean_activity = mean_activity_dict[data_key] if mean_activity_dict is not None else None

            readout_kwargs = self.prepare_readout_kwargs(i, data_key, first_data_key, **kwargs)

            if i == 0 or clone_readout is False:
                self.add_module(
                    data_key,
                    self._base_readout(
                        in_shape=in_shape_dict[data_key],
                        outdims=n_neurons_dict[data_key],
                        mean_activity=mean_activity,
                        **readout_kwargs
                    ),
                )
                original_readout = data_key
            elif i > 0 and clone_readout is True:
                self.add_module(data_key, ClonedReadout(self[original_readout]))

        self.initialize(mean_activity_dict)

    def prepare_readout_kwargs(self, i, data_key, first_data_key, **kwargs):
        return kwargs

    def forward(self, *args, data_key=None, **kwargs):
        if data_key is None and len(self) == 1:
            data_key = list(self.keys())[0]
        return self[data_key](*args, **kwargs)

    def initialize(self, mean_activity_dict=None):
        for data_key, readout in self.items():
            mean_activity = mean_activity_dict[data_key] if mean_activity_dict is not None else None
            readout.initialize(mean_activity)

    def regularizer(self, data_key=None, reduction="sum", average=None):
        if data_key is None and len(self) == 1:
            data_key = list(self.keys())[0]
        return self[data_key].regularizer(reduction=reduction, average=average)


class MultiReadoutSharedParametersBase(MultiReadoutBase):
    """
    Base class for MultiReadouts that share parameters between readouts.
    For more information on which parameters can be shared, refer for example to the FullGaussian2d readout
    """

    def prepare_readout_kwargs(
        self,
        i,
        data_key,
        first_data_key,
        grid_mean_predictor=None,
        grid_mean_predictor_type=None,
        share_transform=False,
        share_grid=False,
        share_features=False,
        **kwargs
    ):
        readout_kwargs = kwargs.copy()

        if grid_mean_predictor:
            if grid_mean_predictor_type == "cortex":
                readout_kwargs["source_grid"] = readout_kwargs["source_grids"][data_key]
                readout_kwargs["grid_mean_predictor"] = grid_mean_predictor
            else:
                raise KeyError("grid mean predictor {} does not exist".format(grid_mean_predictor_type))
            if share_transform:
                readout_kwargs["shared_transform"] = None if i == 0 else self[first_data_key].mu_transform

        elif share_grid:
            readout_kwargs["shared_grid"] = {
                "match_ids": readout_kwargs["shared_match_ids"][data_key],
                "shared_grid": None if i == 0 else self[first_data_key].shared_grid,
            }

        if share_features:
            readout_kwargs["shared_features"] = {
                "match_ids": readout_kwargs["shared_match_ids"][data_key],
                "shared_features": None if i == 0 else self[first_data_key].shared_features,
            }
        else:
            readout_kwargs["shared_features"] = None
        return readout_kwargs



import logging
import warnings

import numpy as np
import torch
from torch import nn
from torch.nn import Parameter
from torch.nn import functional as F

from .base import Readout

logger = logging.getLogger(__name__)


class PointPooled2d(Readout):
    def __init__(
        self,
        in_shape,
        outdims,
        pool_steps,
        bias,
        pool_kern,
        init_range,
        align_corners=True,
        mean_activity=None,
        feature_reg_weight=1.0,
        gamma_readout=None,  # depricated, use feature_reg_weight instead
        **kwargs,
    ):
        """
        This readout learns a point in the core feature space for each neuron, with help of torch.grid_sample, that best
        predicts its response. Multiple average pooling steps are applied to reduce search space in each stage and thereby, faster convergence to the best prediction point.

        The readout receives the shape of the core as 'in_shape', number of pooling stages to be performed as 'pool_steps', the kernel size and stride length
        to be used for pooling as 'pool_kern', the number of units/neurons being predicted as 'outdims', 'bias' specifying whether
        or not bias term is to be used and 'init_range' range for initialising the grid with uniform distribution, U(-init_range,init_range).
        The grid parameter contains the normalized locations (x, y coordinates in the core feature space) and is clipped to [-1.1] as it a
        requirement of the torch.grid_sample function. The feature parameter learns the best linear mapping from the pooled feature
        map from a given location to a unit's response with or without an additional elu non-linearity.

        Args:
            in_shape (list): shape of the input feature map [channels, width, height]
            outdims (int): number of output units
            pool_steps (int): number of pooling stages
            bias (bool): adds a bias term
            pool_kern (int): filter size and stride length used for pooling the feature map
            init_range (float): intialises the grid with Uniform([-init_range, init_range])
                                [expected: positive value <=1]
            align_corners (bool): Keyword agrument to gridsample for bilinear interpolation.
                It changed behavior in PyTorch 1.3. The default of align_corners = True is setting the
                behavior to pre PyTorch 1.3 functionality for comparability.
        """
        super().__init__()
        if init_range > 1.0 or init_range <= 0.0:
            raise ValueError("init_range is not within required limit!")
        self._pool_steps = pool_steps
        self.in_shape = in_shape
        c, w, h = in_shape
        self.outdims = outdims
        self.feature_reg_weight = self.resolve_deprecated_gamma_readout(feature_reg_weight, gamma_readout)
        self.mean_activity = mean_activity
        self.grid = Parameter(torch.Tensor(1, outdims, 1, 2))  # x-y coordinates for each neuron
        self.features = Parameter(
            torch.Tensor(1, c * (self._pool_steps + 1), 1, outdims)
        )  # weight matrix mapping the core features to the output units

        if bias:
            bias = Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias)
        else:
            self.register_parameter("bias", None)

        self.pool_kern = pool_kern
        self.avg = nn.AvgPool2d(
            (pool_kern, pool_kern), stride=pool_kern, count_include_pad=False
        )  # setup kernel of size=[pool_kern,pool_kern] with stride=pool_kern
        self.init_range = init_range
        self.align_corners = align_corners
        self.initialize(mean_activity)

    @property
    def pool_steps(self):
        return self._pool_steps

    @pool_steps.setter
    def pool_steps(self, value):
        assert value >= 0 and int(value) - value == 0, "new pool steps must be a non-negative integer"
        if value != self._pool_steps:
            logger.info("Resizing readout features")
            c, w, h = self.in_shape
            self._pool_steps = int(value)
            self.features = Parameter(torch.Tensor(1, c * (self._pool_steps + 1), 1, self.outdims))
            self.features.data.fill_(1 / self.in_shape[0])

    def initialize(self, mean_activity=None):
        """
        Initialize function initialises the grid, features or weights and bias terms.
        """
        if mean_activity is None:
            mean_activity = self.mean_activity
        self.grid.data.uniform_(-self.init_range, self.init_range)
        self.features.data.fill_(1 / self.in_shape[0])
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)

    def feature_l1(self, reduction="sum", average=None):
        """
        Returns l1 regularization term for features.
        Args:
            average(bool): Deprecated (see reduction) if True, use mean of weights for regularization
            reduction(str): Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'
        """
        return self.apply_reduction(self.features.abs(), reduction=reduction, average=average)

    def regularizer(self, reduction="sum", average=None):
        return self.feature_l1(reduction=reduction, average=average) * self.feature_reg_weight

    def forward(self, x, shift=None, out_idx=None, **kwargs):
        """
        Propagates the input forwards through the readout
        Args:
            x: input data
            shift: shifts the location of the grid (from eye-tracking data)
            out_idx: index of neurons to be predicted

        Returns:
            y: neuronal activity
        """
        self.grid.data = torch.clamp(self.grid.data, -1, 1)
        N, c, w, h = x.size()
        c_in, w_in, h_in = self.in_shape
        if [c_in, w_in, h_in] != [c, w, h]:
            raise ValueError("the specified feature map dimension is not the readout's expected input dimension")

        m = self.pool_steps + 1  # the input feature is considered the first pooling stage
        feat = self.features.view(1, m * c, self.outdims)
        if out_idx is None:
            grid = self.grid
            bias = self.bias
            outdims = self.outdims
        else:
            if isinstance(out_idx, np.ndarray):
                if out_idx.dtype == bool:
                    out_idx = np.where(out_idx)[0]
            feat = feat[:, :, out_idx]
            grid = self.grid[:, out_idx]
            if self.bias is not None:
                bias = self.bias[out_idx]
            outdims = len(out_idx)

        if shift is None:
            grid = grid.expand(N, outdims, 1, 2)
        else:
            # shift grid based on shifter network's prediction
            grid = grid.expand(N, outdims, 1, 2) + shift[:, None, None, :]

        pools = [F.grid_sample(x, grid, align_corners=self.align_corners)]
        for _ in range(self.pool_steps):
            _, _, w_pool, h_pool = x.size()
            if w_pool * h_pool == 1:
                warnings.warn("redundant pooling steps: pooled feature map size is already 1X1, consider reducing it")
            x = self.avg(x)
            pools.append(F.grid_sample(x, grid, align_corners=self.align_corners))
        y = torch.cat(pools, dim=1)
        y = (y.squeeze(-1) * feat).sum(1).view(N, outdims)

        if self.bias is not None:
            y = y + bias
        return y

    def __repr__(self):
        c, w, h = self.in_shape
        r = self.__class__.__name__ + " (" + "{} x {} x {}".format(c, w, h) + " -> " + str(self.outdims) + ")"
        if self.bias is not None:
            r += " with bias"
        r += " and pooling for {} steps\n".format(self.pool_steps)
        for ch in self.children():
            r += "  -> " + ch.__repr__() + "\n"
        return r


class SpatialTransformerPooled3d(Readout):
    def __init__(
        self,
        in_shape,
        outdims,
        pool_steps=1,
        positive=False,
        bias=True,
        init_range=0.05,
        kernel_size=2,
        stride=2,
        grid=None,
        stop_grad=False,
        align_corners=True,
        mean_activity=None,
        feature_reg_weight=1.0,
        gamma_readout=None,  # depricated, use feature_reg_weight instead
    ):
        super().__init__()
        self._pool_steps = pool_steps
        self.in_shape = in_shape
        c, t, w, h = in_shape
        self.outdims = outdims
        self.positive = positive
        self.feature_reg_weight = self.resolve_deprecated_gamma_readout(feature_reg_weight, gamma_readout)
        self.mean_activity = mean_activity
        if grid is None:
            self.grid = Parameter(torch.Tensor(1, outdims, 1, 2))
        else:
            self.grid = grid
        self.features = Parameter(torch.Tensor(1, c * (self._pool_steps + 1), 1, outdims))
        self.register_buffer("mask", torch.ones_like(self.features))

        if bias:
            bias = Parameter(torch.Tensor(outdims))
            self.register_parameter("bias", bias)
        else:
            self.register_parameter("bias", None)

        self.avg = nn.AvgPool2d(kernel_size, stride=stride, count_include_pad=False)
        self.init_range = init_range
        self.initialize(mean_activity)
        self.stop_grad = stop_grad
        self.align_corners = align_corners

    @property
    def pool_steps(self):
        return self._pool_steps

    @pool_steps.setter
    def pool_steps(self, value):
        assert value >= 0 and int(value) - value == 0, "new pool steps must be a non-negative integer"
        if value != self._pool_steps:
            logger.info("Resizing readout features")
            c, t, w, h = self.in_shape
            outdims = self.outdims
            self._pool_steps = int(value)
            self.features = Parameter(torch.Tensor(1, c * (self._pool_steps + 1), 1, outdims))
            self.mask = torch.ones_like(self.features)
            self.features.data.fill_(1 / self.in_shape[0])

    def initialize(self, init_noise=1e-3, grid=True, mean_activity=None):
        if mean_activity is None:
            mean_activity = self.mean_activity
        # randomly pick centers within the spatial map
        self.features.data.fill_(1 / self.in_shape[0])
        if grid:
            self.grid.data.uniform_(-self.init_range, self.init_range)
        if self.bias is not None:
            self.initialize_bias(mean_activity=mean_activity)

    def feature_l1(self, reduction="sum", average=None, subs_idx=None):
        subs_idx = subs_idx if subs_idx is not None else slice(None)
        return self.apply_reduction(self.features[..., subs_idx].abs(), reduction=reduction, average=average)

    def regularizer(self, reduction="sum", average=None):
        return self.feature_l1(reduction=reduction, average=average) * self.feature_reg_weight

    def reset_fisher_prune_scores(self):
        self._prune_n = 0
        self._prune_scores = self.features.detach() * 0

    def update_fisher_prune_scores(self):
        self._prune_n += 1
        if self.features.grad is None:
            raise ValueError("You need to run backward first")
        self._prune_scores += (0.5 * self.features.grad.pow(2) * self.features.pow(2)).detach()

    @property
    def fisher_prune_scores(self):
        return self._prune_scores / self._prune_n

    def prune(self):
        idx = (self.fisher_prune_scores + 1e6 * (1 - self.mask)).squeeze().argmin(dim=0)
        nt = idx.new
        seq = nt(np.arange(len(idx)))
        self.mask[:, idx, :, seq] = 0
        self.features.data[:, idx, :, seq] = 0

    def forward(self, x, shift=None, subs_idx=None, **kwargs):
        if self.stop_grad:
            x = x.detach()

        self.features.data *= self.mask

        if self.positive:
            self.features.data.clamp_min_(0)
        self.grid.data = torch.clamp(self.grid.data, -1, 1)

        N, c, t, w, h = x.size()
        m = self._pool_steps + 1
        if subs_idx is not None:
            feat = self.features[..., subs_idx].contiguous()
            outdims = feat.size(-1)
            feat = feat.view(1, m * c, outdims)
            grid = self.grid[:, subs_idx, ...]
        else:
            grid = self.grid
            feat = self.features.view(1, m * c, self.outdims)
            outdims = self.outdims

        if shift is None:
            grid = grid.expand(N * t, outdims, 1, 2)
        else:
            grid = grid.expand(N, outdims, 1, 2)
            grid = torch.stack([grid + shift[:, i, :][:, None, None, :] for i in range(t)], 1)
            grid = grid.contiguous().view(-1, outdims, 1, 2)
        z = x.contiguous().transpose(2, 1).contiguous().view(-1, c, w, h)
        pools = [F.grid_sample(z, grid, align_corners=self.align_corners)]
        for i in range(self._pool_steps):
            z = self.avg(z)
            pools.append(F.grid_sample(z, grid, align_corners=self.align_corners))
        y = torch.cat(pools, dim=1)
        y = (y.squeeze(-1) * feat).sum(1).view(N, t, outdims)

        if self.bias is not None:
            if subs_idx is None:
                y = y + self.bias
            else:
                y = y + self.bias[subs_idx]

        return y

    def __repr__(self):
        c, _, w, h = self.in_shape
        r = self.__class__.__name__ + " (" + "{} x {} x {}".format(c, w, h) + " -> " + str(self.outdims) + ")"
        if self.bias is not None:
            r += " with bias"
        if self.stop_grad:
            r += ", stop_grad=True"
        r += "\n"

        for ch in self.children():
            r += "  -> " + ch.__repr__() + "\n"
        return r
